#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def match_phibe_v(b_0, b_z, int_e_dt, n, v):
    r"""Get propagation velocity by matching dBpar and phi. Used together with
    irf_match_phibe_dir.m.Finds best match in amplitude given, B0, dB_par,
    phi, propagation direction implied, for specified n and v given as
    vectors.Returns a matrix of correlations and the two potentials that were
    correlated.

    Parameters
    ----------
    b_0 : array_like
        Average background magnetic field.
    b_z : array_like
        Parallel wave magnetic field.
    int_e_dt : array_like
        Potential.
    n : array_like
        Vector of densities
    v : array_like
        Vector of velocities.

    Returns
    -------
    corr_mat : numpy.ndarray
        Correlation matrix(nn x nv).
    phi_b : numpy.ndarray
        B0 * dB_par / n_e * e * mu0
    phi_e : numpy.ndarray
        int(E) dt * v(dl=-vdt = > -dl = vdt)

    """

    # Define constants
    mu0 = constants.mu_0
    q_e = constants.elementary_charge

    # density in #/m^3
    n.data *= 1e6

    # Allocate correlations matrix rows: n, cols: v
    nn_, nv_ = [len(n), len(v)]
    corr_mat = np.zeros((nn_, nv_))

    # Setup potentials
    phi_e = int_e_dt * v  # depends on v
    phi_b = np.transpose(b_z[:, 0] * b_0 * 1e-18 / (mu0 * q_e * n[:, None]))

    # Get correlation
    for k, p in itertools.product(range(nn_), range(nv_)):
        corr_mat[k, p] = np.sum((np.log10(abs(phi_e[:, p]) / phi_b[:, k])))

    return corr_mat, phi_b, phi_e



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import itertools
import json
import os
import random
import string

# Built-in imports
import unittest

# 3rd party imports
import numpy as np
import requests
import xarray as xr
from ddt import data, ddt, idata, unpack

from pyrfu import mms, pyrf
from pyrfu.mms.psd_moments import _moms

from . import (
    generate_data,
    generate_spectr,
    generate_timeline,
    generate_ts,
    generate_vdf,
)

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

TEST_TINT = ["2019-01-01T00:00:00.000000000", "2019-01-01T00:10:00.000000000"]


def generate_feeps(f_s, n_pts, data_rate, dtype, lev, mms_id):
    var = {"tmmode": data_rate, "dtype": dtype, "lev": lev}
    eyes = mms.feeps_active_eyes(var, TEST_TINT, mms_id)
    keys = [f"{k}-{eyes[k][i]}" for k in eyes for i in range(len(eyes[k]))]
    feeps_dict = {k: generate_spectr(f_s, n_pts, 16, f"energy-{k}") for k in keys}
    feeps_dict["spinsectnum"] = pyrf.ts_scalar(
        generate_timeline(f_s, n_pts), np.tile(np.arange(12), n_pts // 12 + 1)[:n_pts]
    )

    feeps_alle = xr.Dataset(feeps_dict)
    feeps_alle.attrs = {"mmsId": mms_id, **var}

    return feeps_alle


def generate_eis(f_s, n_pts, data_rate, dtype, lev, specie, data_unit, mms_id):
    pref = f"mms{mms_id:d}_epd_eis"
    pref = f"{pref}_{data_rate}_{lev}_{dtype}"

    if data_rate == "brst":
        pref = f"{pref}_{data_rate}_{dtype}"
    else:
        pref = f"{pref}_{dtype}"

    suf = f"{specie}_P1_{data_unit.lower()}_t"

    keys = [f"{pref}_{suf}{t:d}" for t in range(6)]

    spin_nums = pyrf.ts_scalar(
        generate_timeline(f_s, n_pts),
        np.sort(np.tile(np.arange(n_pts // 12 + 1), (12,)))[1 : n_pts + 1],
    )
    sectors = pyrf.ts_scalar(
        generate_timeline(f_s, n_pts),
        np.tile(np.arange(12), n_pts // 12 + 1)[1 : n_pts + 1],
    )

    if dtype.lower() == "extof":
        energies = np.array(
            [
                47.645324,
                54.928681,
                62.419454,
                70.833554,
                80.315371,
                91.00098,
                103.018894,
                116.554129,
                131.801143,
                148.970297,
                168.295534,
                190.060874,
                214.590996,
                242.245343,
                273.466432,
                308.768669,
                348.73539,
                394.035378,
                445.404668,
                503.597543,
                569.429005,
                643.764143,
                727.683404,
                822.660211,
                930.654627,
            ]
        )
    else:
        energies = np.array(
            [
                10.51516,
                11.509144,
                12.612351,
                13.817409,
                15.111664,
                16.55435,
                18.134081,
                19.857029,
                21.774935,
                23.807037,
                26.021971,
                28.526016,
                31.215776,
                34.228877,
                37.604494,
                41.116729,
                45.29041,
                51.412368,
                58.570702,
                65.951929,
                75.09237,
            ]
        )

    eis_dict = {"spin": spin_nums, "sector": sectors}

    for i, k in enumerate(keys):
        eis_dict[f"t{i:d}"] = generate_spectr(f_s, n_pts, len(energies), "energy")
        eis_dict[f"look_t{i:d}"] = generate_ts(f_s, n_pts, "vector")

    # glob_attrs = {**outdict["spin"].attrs["GLOBAL"], **var}
    glob_attrs = {
        "delta_energy_plus": 0.5 * np.ones(len(energies)),
        "delta_energy_minus": 0.5 * np.ones(len(energies)),
        "species": specie,
        "randattrs": "".join(random.choice(string.ascii_lowercase) for _ in range(10)),
    }

    # Build Dataset
    eis = xr.Dataset(eis_dict, attrs=glob_attrs)
    eis = eis.assign_coords(energy=energies)

    return eis


def _mms_keys():
    test_path = os.path.dirname(os.path.abspath(__file__))

    root_path = os.path.join(os.path.split(test_path)[0], "mms")

    with open(
        os.sep.join([root_path, "mms_keys.json"]), "r", encoding="utf-8"
    ) as json_file:
        keys_ = json.load(json_file)

    all_keys = list(
        np.hstack([list(instrument.keys()) for instrument in keys_.values()])
    )
    return all_keys


@ddt
class CalcEpsilonTestCase(unittest.TestCase):
    @data(
        (
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="bazinga"),
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="bazinga"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
        ),
        (
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="ions"),
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="ions"),
            generate_ts(32.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
        ),
    )
    @unpack
    def test_calc_epsilon_input(self, vdf, model_vdf, n_s, sc_pot):
        with self.assertRaises(ValueError):
            mms.calculate_epsilon(vdf, model_vdf, n_s, sc_pot)

    @data(
        (
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="ions"),
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="ions"),
            {},
        ),
        (
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="electrons"),
            generate_vdf(64.0, 100, (32, 16, 16), energy01=False, species="electrons"),
            {},
        ),
        (
            generate_vdf(64.0, 100, (32, 16, 16), energy01=True, species="ions"),
            generate_vdf(64.0, 100, (32, 16, 16), energy01=True, species="ions"),
            {},
        ),
    )
    @unpack
    def test_calc_epsilon_output(self, vdf, model_vdf, kwargs):
        mms.calculate_epsilon(
            vdf,
            model_vdf,
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
            **kwargs,
        )


class DbInitTestCase(unittest.TestCase):
    def test_db_init_inpput(self):
        with self.assertRaises(AssertionError):
            mms.db_init("bazinga!")

    def test_db_init_output(self):
        self.assertIsNone(mms.db_init(os.getcwd()))


@ddt
class Def2PsdTestCase(unittest.TestCase):
    @data(
        ("I AM GROOT!!", "s^3/cm^6"),
        ("ions", "bazinga"),
    )
    @unpack
    def test_def2psd_input(self, species, units):
        with self.assertRaises(ValueError):
            mms.def2psd(generate_vdf(64.0, 100, (32, 32, 16), False, species, units))

    @idata(
        itertools.product(
            [
                "ions",
                "ion",
                "protons",
                "proton",
                "alphas",
                "alpha",
                "helium",
                "electrons",
                "e",
            ],
            ["keV/(cm^2 s sr keV)", "eV/(cm^2 s sr eV)", "1/(cm^2 s sr)"],
        )
    )
    @unpack
    def test_def2psd_output(self, species, units):
        vdf = generate_vdf(64.0, 100, (32, 32, 16), False, species, units)
        result = mms.def2psd(vdf)
        self.assertIsInstance(result, xr.Dataset)

        spectr = generate_spectr(64.0, 100, 32, {"species": species, "UNITS": units})
        result = mms.def2psd(spectr)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class Dpf2PsdTestCase(unittest.TestCase):
    @data(
        ("I AM GROOT!!", "s^3/cm^6"),
        ("ions", "bazinga"),
    )
    @unpack
    def test_dpf2psd_input(self, species, units):
        with self.assertRaises(ValueError):
            mms.dpf2psd(generate_vdf(64.0, 100, (32, 32, 16), False, species, units))

    @idata(
        itertools.product(
            [
                "ions",
                "ion",
                "protons",
                "proton",
                "alphas",
                "alpha",
                "helium",
                "electrons",
                "e",
            ],
            ["1/(cm^2 s sr keV)", "1/(cm^2 s sr eV)"],
        )
    )
    @unpack
    def test_dpf2psd_output(self, species, units):
        vdf = generate_vdf(64.0, 100, (32, 32, 16), False, species, units)
        result = mms.dpf2psd(vdf)
        self.assertIsInstance(result, xr.Dataset)

        spectr = generate_spectr(64.0, 100, 32, {"species": species, "UNITS": units})
        result = mms.dpf2psd(spectr)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class Dsl2GseTestCase(unittest.TestCase):
    def test_dsl2gse_input(self):
        with self.assertRaises(TypeError):
            mms.dsl2gse(generate_ts(64.0, 42, "vector"), np.random.random((42, 3)), 1)

    @data(
        xr.Dataset({"z_dec": generate_ts(64.0, 42), "z_ra": generate_ts(64.0, 42)}),
        np.random.random(3),
    )
    def test_dsl2gse_output(self, value):
        result = mms.dsl2gse(generate_ts(64.0, 42, "vector"), value, 1)
        self.assertIsInstance(result, xr.DataArray)
        result = mms.dsl2gse(generate_ts(64.0, 42, "vector"), value, -1)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class Dsl2GsmTestCase(unittest.TestCase):
    def test_dsl2gsm_input(self):
        with self.assertRaises(TypeError):
            mms.dsl2gsm(generate_ts(64.0, 42, "vector"), np.random.random((42, 3)), 1)

    @data(
        xr.Dataset({"z_dec": generate_ts(64.0, 42), "z_ra": generate_ts(64.0, 42)}),
        np.random.random(3),
    )
    def test_dsl2gsm_output(self, value):
        result = mms.dsl2gsm(generate_ts(64.0, 42, "vector"), value, 1)
        self.assertIsInstance(result, xr.DataArray)
        result = mms.dsl2gsm(generate_ts(64.0, 42, "vector"), value, -1)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EisCombineProtonPadTestCase(unittest.TestCase):
    @idata(
        itertools.product(
            ["srvy", "brst"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_combine_proton_pad_output(self, tmmode, specie, unit):
        mms_id = random.randint(1, 5)
        phxtof_allt = generate_eis(
            64.0, 100, tmmode, "phxtof", "l2", specie, unit, mms_id
        )
        extof_allt = generate_eis(
            64.0, 100, tmmode, "extof", "l2", specie, unit, mms_id
        )
        result = mms.eis_combine_proton_pad(phxtof_allt, extof_allt)
        self.assertIsInstance(result, xr.DataArray)

    @idata(itertools.product([99, 100], repeat=2))
    @unpack
    def test_eis_combine_proton_pad_input(self, n_phxtof, n_extof):
        phxtof_allt = generate_eis(
            64.0, n_phxtof, "brst", "phxtof", "l2", "proton", "flux", 1
        )
        extof_allt = generate_eis(
            64.0, n_extof, "brst", "extof", "l2", "proton", "flux", 1
        )
        result = mms.eis_combine_proton_pad(phxtof_allt, extof_allt)
        self.assertIsInstance(result, xr.DataArray)

    @data(None, [1, 0, 0], generate_ts(64.0, 10, "vector"))
    def test_eis_combine_proton_pad_vec(self, vec):
        phxtof_allt = generate_eis(
            64.0, 100, "brst", "phxtof", "l2", "proton", "flux", 1
        )
        extof_allt = generate_eis(64.0, 100, "brst", "extof", "l2", "proton", "flux", 1)
        result = mms.eis_combine_proton_pad(phxtof_allt, extof_allt, vec)
        self.assertIsInstance(result, xr.DataArray)

    def test_eis_combine_proton_pad_options(self):
        phxtof_allt = generate_eis(
            64.0, 100, "brst", "phxtof", "l2", "proton", "flux", 1
        )
        extof_allt = generate_eis(64.0, 100, "brst", "extof", "l2", "proton", "flux", 1)
        result = mms.eis_combine_proton_pad(phxtof_allt, extof_allt, None, despin=True)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EisCombineProtonSpecTestCase(unittest.TestCase):
    @idata(
        itertools.product(
            ["srvy", "brst"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_combine_proton_spec_output(self, tmmode, specie, unit):
        mms_id = random.randint(1, 5)
        phxtof_allt = generate_eis(
            64.0, 100, tmmode, "phxtof", "l2", specie, unit, mms_id
        )
        extof_allt = generate_eis(
            64.0, 100, tmmode, "extof", "l2", specie, unit, mms_id
        )
        result = mms.eis_combine_proton_spec(phxtof_allt, extof_allt)
        self.assertIsInstance(result, xr.Dataset)

    @idata(itertools.product([99, 100], repeat=2))
    @unpack
    def test_eis_combine_proton_spec_ctimes(self, n_phxtof, n_extof):
        phxtof_allt = generate_eis(
            64.0, n_phxtof, "brst", "phxtof", "l2", "proton", "flux", 1
        )
        extof_allt = generate_eis(
            64.0, n_extof, "brst", "extof", "l2", "proton", "flux", 1
        )
        result = mms.eis_combine_proton_spec(phxtof_allt, extof_allt)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class EisOmniTestCase(unittest.TestCase):
    @idata(
        itertools.product(
            ["srvy", "brst"],
            ["extof", "phxtof"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_omni_output(self, tmmode, dtype, specie, unit):
        eis = generate_eis(
            64.0, 100, tmmode, dtype, "l2", specie, unit, random.randint(1, 4)
        )
        result = mms.eis_omni(eis, "mean")
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EisPadTestCase(unittest.TestCase):
    @data(None, [1, 0, 0], generate_ts(64.0, 10, "vector"))
    def test_eis_pad_output(self, vec):
        eis = generate_eis(64.0, 100, "brst", "extof", "l2", "proton", "flux", 1)
        result = mms.eis_pad(eis, vec)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EisPadSpinAvgTestCase(unittest.TestCase):
    @idata(
        itertools.product(
            ["srvy", "brst"],
            ["extof", "phxtof"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_pad_spin_avg_output(self, tmmode, dtype, specie, unit):
        eis = generate_eis(
            64.0, 100, tmmode, dtype, "l2", specie, unit, random.randint(1, 4)
        )
        eis_pad = mms.eis_pad(eis)
        result = mms.eis_pad_spinavg(eis_pad, eis.spin)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class EisProtonCorrectionTestCase(unittest.TestCase):
    def test_eis_proton_correction_dataarray(self):
        flux_eis = generate_spectr(64, 100, 16, "energy")
        result = mms.eis_proton_correction(flux_eis)
        self.assertIsInstance(result, xr.DataArray)

    @idata(
        itertools.product(
            ["srvy", "brst"],
            ["extof", "phxtof"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_proton_correction_dataset(self, tmmode, dtype, specie, unit):
        flux_eis = generate_eis(
            64.0, 100, tmmode, dtype, "l2", specie, unit, random.randint(1, 4)
        )
        result = mms.eis_proton_correction(flux_eis)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class EisSpinAvgTestCase(unittest.TestCase):
    @idata(
        itertools.product(
            ["mean", "sum"],
            ["srvy", "brst"],
            ["extof", "phxtof"],
            ["proton", "alpha", "oxygen"],
            ["flux", "cps", "counts"],
        )
    )
    @unpack
    def test_eis_spin_avg_output(self, method, tmmode, dtype, specie, unit):
        eis_allt = generate_eis(
            64.0, 100, tmmode, dtype, "l2", specie, unit, random.randint(1, 4)
        )
        result = mms.eis_spin_avg(eis_allt, method)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class MakeModelVDFTestCase(unittest.TestCase):
    @data(
        (generate_vdf(64.0, 100, (32, 16, 16), species="ions"), False),
        (generate_vdf(64.0, 100, (32, 16, 16), species="electrons"), False),
        (generate_vdf(64.0, 100, (32, 16, 16), species="ions"), True),
    )
    @unpack
    def test_make_Model_vdf_output(self, vdf, isotropic):
        result = mms.make_model_vdf(
            vdf,
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "tensor"),
            isotropic,
        )
        self.assertIsInstance(result, xr.Dataset)


class HpcaEnergiesTestCase(unittest.TestCase):
    def test_hpca_energies_output(self):
        result = mms.hpca_energies()
        self.assertIsInstance(result, list)


@ddt
class MakeModelKappaTestCase(unittest.TestCase):
    @data(
        (generate_vdf(64.0, 100, (32, 16, 16), species="bazinga"), random.random()),
    )
    @unpack
    def test_make_model_kappa_input(self, vdf, kappa):
        with self.assertRaises(ValueError):
            mms.make_model_kappa(
                vdf,
                generate_ts(64.0, 100, "scalar"),
                generate_ts(64.0, 100, "vector"),
                generate_ts(64.0, 100, "scalar"),
                kappa,
            )

    @data(
        (generate_vdf(64.0, 100, (32, 16, 16), species="ions"), random.random()),
        (generate_vdf(64.0, 100, (32, 16, 16), species="electrons"), random.random()),
        (generate_vdf(64.0, 100, (32, 16, 16), species="ions"), random.random()),
    )
    @unpack
    def test_make_model_kappa_output(self, vdf, kappa):
        result = mms.make_model_kappa(
            vdf,
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "vector"),
            generate_ts(64.0, 100, "scalar"),
            kappa,
        )

        self.assertIsInstance(result, xr.Dataset)


@ddt
class Psd2DefTestCase(unittest.TestCase):
    @data(
        ("I AM GROOT!!", "s^3/cm^6"),
        ("ions", "bazinga"),
    )
    @unpack
    def test_psd2def_input(self, species, units):
        with self.assertRaises(ValueError):
            mms.psd2def(generate_vdf(64.0, 100, (32, 32, 16), False, species, units))

    @idata(
        itertools.product(
            [
                "ions",
                "ion",
                "protons",
                "proton",
                "alphas",
                "alpha",
                "helium",
                "electrons",
                "e",
            ],
            ["s^3/cm^6", "s^3/m^6", "s^3/km^6"],
        )
    )
    @unpack
    def test_psd2def_output(self, species, units):
        vdf = generate_vdf(64.0, 100, (32, 32, 16), False, species, units)
        result = mms.psd2def(vdf)
        self.assertIsInstance(result, xr.Dataset)

        spectr = generate_spectr(64.0, 100, 32, {"species": species, "UNITS": units})
        result = mms.psd2def(spectr)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class Psd2DpfTestCase(unittest.TestCase):
    @data(
        ("I AM GROOT!!", "s^3/cm^6"),
        ("ions", "bazinga"),
    )
    @unpack
    def test_psd2dpf_input(self, species, units):
        with self.assertRaises(ValueError):
            mms.psd2dpf(generate_vdf(64.0, 100, (32, 32, 16), False, species, units))

    @idata(
        itertools.product(
            [
                "ions",
                "ion",
                "protons",
                "proton",
                "alphas",
                "alpha",
                "helium",
                "electrons",
                "e",
            ],
            ["s^3/cm^6", "s^3/m^6", "s^3/km^6"],
        )
    )
    @unpack
    def test_psd2dpf_output(self, species, units):
        vdf = generate_vdf(64.0, 100, (32, 32, 16), False, species, units)
        result = mms.psd2dpf(vdf)
        self.assertIsInstance(result, xr.Dataset)

        spectr = generate_spectr(64.0, 100, 32, {"species": species, "UNITS": units})
        result = mms.psd2dpf(spectr)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class PsdMomentsTestCase(unittest.TestCase):
    @data(
        (generate_vdf(64.0, 100, (32, 32, 16)), "brst"),
        (generate_vdf(64.0, 100, (32, 32, 16), species="electrons"), "brst"),
        (generate_vdf(64.0, 100, (32, 32, 16), energy01=False), "brst"),
        (generate_vdf(64.0, 100, (32, 32, 16), energy01=False), "fast"),
        (generate_vdf(64.0, 100, (32, 32, 16), energy01=True), "brst"),
    )
    @unpack
    def test_psd_moments_input(self, vdf, data_rate):
        delta_theta = 0.5 * np.ones(vdf.data.shape[3])
        vdf.attrs["delta_theta_minus"] = delta_theta
        vdf.attrs["delta_theta_plus"] = delta_theta

        delta_phi = 0.5 * np.ones((vdf.data.shape[0], vdf.data.shape[2]))
        vdf.attrs["delta_phi_minus"] = delta_phi
        vdf.attrs["delta_phi_plus"] = delta_phi
        vdf.data.attrs["FIELDNAM"] = f"MMS1 FPI/DIS {data_rate}SkyMap dist"
        mms.psd_moments(vdf, generate_ts(64.0, 100, "scalar"))

    @data({"energy_range": [1, 1000]}, {"no_sc_pot": True})
    def test_psd_moments_options(self, options):
        vdf = generate_vdf(64.0, 100, (32, 32, 16))
        vdf.data.attrs["FIELDNAM"] = "MMS1 FPI/DIS brstSkyMap dist"
        mms.psd_moments(vdf, generate_ts(64.0, 100, "scalar"), **options)

    @data(
        (
            np.random.random((100, 32)),  # energy
            np.random.random((10000, 32)),  # delta_v
            random.random(),  # q_e
            np.random.random(100),  # sc_pot
            random.random(),  # p_mass
            random.choice([True, False]),  # flag_inner_electron
            random.random(),  # w_inner_electron
            np.random.random((100, 32, 16)),  # phi
            np.random.random((100, 32, 16)),  # theta
            np.arange(32),  # int_energies
            np.random.random((100, 32, 32, 16)),  # vdf
            np.random.random((100, 32, 16)),  # delta_ang
        )
    )
    def test_moms(self, value):
        result = _moms.__wrapped__(*value)
        self.assertIsInstance(result[0], np.ndarray)
        self.assertIsInstance(result[1], np.ndarray)
        self.assertIsInstance(result[2], np.ndarray)
        self.assertIsInstance(result[3], np.ndarray)


@ddt
class PsdRebinTestCase(unittest.TestCase):
    @data(generate_vdf(64.0, 100, (32, 32, 16), energy01=True, species="ions"))
    def test_psd_rebin_output(self, vdf):
        result = mms.psd_rebin(
            vdf,
            vdf.phi,
            vdf.attrs["energy0"],
            vdf.attrs["energy1"],
            vdf.attrs["esteptable"],
        )
        self.assertIsInstance(result[0], np.ndarray)
        self.assertEqual(len(result[0]), 50)
        self.assertIsInstance(result[1], np.ndarray)
        self.assertListEqual(list(result[1].shape), [50, 64, 32, 16])
        self.assertIsInstance(result[2], np.ndarray)
        self.assertEqual(len(result[2]), 64)
        self.assertIsInstance(result[3], np.ndarray)
        self.assertListEqual(list(result[3].shape), [50, 32])


@ddt
class FeepsActiveEyesTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"], ["sitl", "l2"]))
    @unpack
    def test_feeps_active_eyes_output(self, data_rate, dtype, lev):
        result = mms.feeps_active_eyes(
            {"tmmode": data_rate, "dtype": dtype, "lev": lev},
            TEST_TINT,
            random.randint(1, 4),
        )
        self.assertIsInstance(result, dict)

        result = mms.feeps_active_eyes(
            {"tmmode": data_rate, "dtype": dtype, "lev": lev},
            pyrf.iso86012datetime64(np.array(TEST_TINT)),
            str(random.randint(1, 4)),
        )
        self.assertIsInstance(result, dict)


@ddt
class FeepsCorrectEnergiesTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_correct_energies_output(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )

        result = mms.feeps_correct_energies(feeps_alle)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class FeepsFlatFieldCorrectionsTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_flat_field_corrections_output(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )

        result = mms.feeps_flat_field_corrections(feeps_alle)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class FeepsOmniTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_omni_output(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)

        result = mms.feeps_omni(feeps_alle)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class FeepsPadTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_pad_ouput(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)

        result = mms.feeps_pad(feeps_alle, generate_ts(64.0, 100, "vector"))
        self.assertIsInstance(result, xr.DataArray)


@ddt
class FeepsPadSpinAvgTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_pad_spin_avg(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)

        feeps_pad = mms.feeps_pad(feeps_alle, generate_ts(64.0, 100, "vector"))
        result = mms.feeps_pad_spinavg(feeps_pad, feeps_alle.spinsectnum)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class FeepsPitchAnglesTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_pitch_angles_output(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)

        result = mms.feeps_pitch_angles(feeps_alle, generate_ts(64.0, 100, "vector"))
        self.assertIsInstance(result[0], xr.DataArray)


@ddt
class FeepsRemoveBadDataTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_remove_bad_data_output(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )

        result = mms.feeps_remove_bad_data(feeps_alle)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class FeepsRemoveSunTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_remove_sun(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)

        result = mms.feeps_remove_sun(feeps_alle)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class FeepsSpinAvgTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_spin_avg(self, data_rate, dtype):
        # Generate fake FEEPS data
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        feeps_alle, _ = mms.feeps_split_integral_ch(feeps_alle)
        feeps_omni = mms.feeps_omni(feeps_alle)

        result = mms.feeps_spin_avg(feeps_omni, feeps_alle.spinsectnum)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class FeepsSplitIntegralChTestCase(unittest.TestCase):
    @idata(itertools.product(["srvy", "brst"], ["electron", "ion"]))
    @unpack
    def test_feeps_split_integral_ch(self, data_rate, dtype):
        feeps_alle = generate_feeps(
            64.0, 100, data_rate, dtype, "l2", random.randint(1, 4)
        )
        mms.feeps_split_integral_ch(feeps_alle)


@ddt
class FkPowerSpectrum4scTestCase(unittest.TestCase):
    @data((None, None), (random.random(), None), (None, [0.1, 1]))
    @unpack
    def test_fk_power_spectrum_4sc(self, df, f_range):
        e_mms = [generate_ts(64.0, 100, "scalar") for _ in range(4)]
        r_mms = [generate_ts(64.0, 100, "vector") for _ in range(4)]
        b_mms = [generate_ts(64.0, 100, "vector") for _ in range(4)]

        result = mms.fk_power_spectrum_4sc(
            e_mms, r_mms, b_mms, TEST_TINT, df=df, f_range=f_range
        )
        self.assertIsInstance(result, xr.Dataset)


@ddt
class ReduceTestCase(unittest.TestCase):
    @data("s^3/cm^6", "s^3/m^6", "s^3/km^6")
    def test_reduce_units(self, value):
        vdf = generate_vdf(64.0, 42, [32, 32, 16], energy01=True, species="ions")
        vdf.data.attrs["UNITS"] = value
        result = mms.reduce(vdf, np.eye(3), "1d", "cart")
        self.assertIsInstance(result, xr.DataArray)

    @data(
        (False, "ions", np.eye(3), "1d", "cart"),
        (False, "electrons", np.eye(3), "1d", "cart"),
        (True, "ions", np.eye(3), "1d", "cart"),
        (True, "electrons", np.eye(3), "1d", "cart"),
        (False, "electrons", generate_ts(64.0, 42, "tensor"), "1d", "cart"),
        (False, "ions", np.eye(3), "1d", "pol"),
        # (False, "ions", np.eye(3), "2d", "pol"), mc_pol_2d NotImplementedError
    )
    @unpack
    def test_reduce_output(self, energy01, species, xyz, dim, base):
        vdf = generate_vdf(64.0, 42, [32, 32, 16], energy01, species)
        result = mms.reduce(vdf, xyz, dim, base)
        self.assertIsInstance(result, xr.DataArray)

    @data(
        ("1d", "cart", {"vg": np.linspace(-1, 1, 42)}),
        ("1d", "cart", {"lower_e_lim": generate_ts(64.0, 42)}),
        ("1d", "cart", {"vg_edges": np.linspace(-1.01, 1.01, 102)}),
    )
    @unpack
    def test_reduce_options(self, dim, base, options):
        vdf = generate_vdf(64.0, 42, [32, 32, 16], energy01=False, species="ions")
        xyz = np.eye(3)
        result = mms.reduce(vdf, xyz, dim, base, **options)
        self.assertIsInstance(result, xr.DataArray)

    @data(
        ("ions", "s^3/m^6", np.array([1, 0, 0]), "1d", "pol", {}),
        ("I AM GROOT", "s^3/m^6", np.eye(3), "1d", "pol", {}),
        ("ions", "bazinga", np.eye(3), "1d", "pol", {}),
        ("ions", "s^3/m^6", np.eye(3), "2d", "pol", {}),
        ("ions", "s^3/m^6", np.eye(3), "1d", "pol", {"lower_e_lim": generate_data(42)}),
    )
    @unpack
    def test_reduce_input(self, species, units, xyz, dim, base, options):
        vdf = generate_vdf(64.0, 42, [32, 32, 16], energy01=True, species=species)
        vdf.data.attrs["UNITS"] = units
        with self.assertRaises((TypeError, ValueError, NotImplementedError)):
            mms.reduce(vdf, xyz, dim, base, **options)


@ddt
class RotateTensorTestCase(unittest.TestCase):
    @data(("rot", generate_data(100, "vector")), ("gse", None))
    @unpack
    def test_rotate_tensor_input(self, flag, vec):
        with self.assertRaises((TypeError, NotImplementedError)):
            mms.rotate_tensor(generate_ts(64.0, 100, "tensor"), flag, vec)

    @data(
        ("fac", generate_ts(64.0, 100, "vector"), "pp"),
        ("fac", generate_ts(64.0, 100, "vector"), "qq"),
        ("rot", np.random.random(3), "pp"),
        ("rot", np.random.random((3, 3)), "pp"),
    )
    @unpack
    def test_rotate_tensor_output(self, rot_flag, vec, perp):
        result = mms.rotate_tensor(
            generate_ts(64.0, 100, "tensor"), rot_flag, vec, perp
        )
        self.assertIsInstance(result, xr.DataArray)


@ddt
class SpectrToDatasetTestCase(unittest.TestCase):
    @data(generate_spectr(64.0, 100, 10))
    def test_spectr_to_dataset_output(self, spectr):
        result = mms.spectr_to_dataset(spectr)
        self.assertIsInstance(result, xr.Dataset)


@ddt
class Scpot2NeTestCase(unittest.TestCase):
    @data(None, generate_ts(64.0, 100, "scalar"))
    def test_scpot2ne_output(self, i_aspoc):
        result = mms.scpot2ne(
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "scalar"),
            generate_ts(64.0, 100, "tensor"),
            i_aspoc,
        )
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], float)
        self.assertIsInstance(result[2], float)
        self.assertIsInstance(result[3], float)
        self.assertIsInstance(result[4], float)


@ddt
class VdfElimTestCase(unittest.TestCase):
    @data(
        random.randint(0, 15),
        random.randint(0, 15) + 0.4,
        [random.randint(0, 15), random.randint(16, 31)],
    )
    def test_vdf_elim_output(self, e_int):
        result = mms.vdf_elim(
            generate_vdf(64.0, 42, [32, 32, 16], energy01=True), e_int
        )
        self.assertIsInstance(result, xr.Dataset)


@ddt
class VdfOmniTestCase(unittest.TestCase):
    @data("mean", "sum")
    def test_vdf_omni_output(self, method):
        result = mms.vdf_omni(generate_vdf(64.0, 100, (32, 32, 16)), method)
        self.assertIsInstance(result, xr.DataArray)


@ddt
class TokenizeTestCase(unittest.TestCase):
    @data(*random.choices(_mms_keys(), k=10))
    def test_tokenize(self, var_str):
        result = mms.tokenize(var_str)
        self.assertIsInstance(result, dict)


@ddt
class ListFilesTestCase(unittest.TestCase):
    @data(*random.choices(_mms_keys(), k=10))
    def test_list_files(self, var_str):
        mms.list_files(TEST_TINT, random.randint(1, 4), mms.tokenize(var_str))


@ddt
class ListFilesSdcTestCase(unittest.TestCase):
    @data(*random.choices(_mms_keys(), k=10))
    def test_list_files_sdc(self, var_str):
        try:
            mms.list_files_sdc(TEST_TINT, random.randint(1, 4), mms.tokenize(var_str))
        except requests.exceptions.ReadTimeout:
            pass


@ddt
class ListFilesAncillaryTestCase(unittest.TestCase):
    @data("predatt", "predeph", "defatt", "defeph")
    def test_list_files_ancillary(self, product):
        mms.list_files_ancillary(TEST_TINT, random.randint(1, 4), product)


@ddt
class ListFilesAncillarySdcTestCase(unittest.TestCase):
    @data("predatt", "predeph", "defatt", "defeph")
    def test_list_files_ancillary_sdc(self, product):
        try:
            mms.list_files_ancillary_sdc(TEST_TINT, random.randint(1, 4), product)
        except requests.exceptions.ReadTimeout:
            pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

import matplotlib.pyplot as plt

# 3rd party imports
import xarray as xr

# Local imports
from ..pyrf import histogram2d
from . import plot_spectr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pl_scatter_matrix(
    inp1,
    inp2: xr.DataArray = None,
    pdf: bool = False,
    cmap: str = "jet",
):
    r"""Produces a scatter plot of each components of field inp1 with respect
    to every component of field inp2. If pdf is set to True, the scatter
    plot becomes a 2d histogram.

    Parameters
    ----------
    inp1 : xarray.DataArray
        First time series (x-axis).
    inp2 : xarray.DataArray
        Second time series (y-axis).
    pdf : bool, Optional
        Flag to plot the 2d histogram. If False the figure is a scatter plot.
        If True the figure is a 2d histogram.
    cmap : str, Optional
        Colormap. Default : "jet"

    Returns
    -------
    fig : matplotlib.pyplot.figure
        Figure with time series plots.
    axs : matplotlib.axes._subplots.AxesSubplot
        Axes.
    caxs : matplotlib.pyplot.colorbar
        Only if pdf is True

    """

    if inp2 is None:
        inp2 = inp1
        warnings.warn("inp2 is empty assuming that inp2=inp1", UserWarning)

    assert isinstance(inp1, xr.DataArray) and isinstance(inp2, xr.DataArray)

    if not pdf:
        fig, axs = plt.subplots(
            3,
            3,
            sharex="all",
            sharey="all",
            figsize=(16, 9),
        )
        fig.subplots_adjust(
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.9,
            hspace=0.05,
            wspace=0.05,
        )

        for i in range(3):
            for j in range(3):
                axs[j, i].scatter(
                    inp1[:, i].data,
                    inp2[:, j].data,
                    marker="+",
                )

        out = (fig, axs)
    else:
        fig, axs = plt.subplots(
            3,
            3,
            sharex="all",
            sharey="all",
            figsize=(16, 9),
        )
        fig.subplots_adjust(
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.9,
            hspace=0.05,
            wspace=0.3,
        )

        caxs = [[None] * 3] * 3

        for i in range(3):
            for j in range(3):
                hist_ = histogram2d(inp1[:, i], inp2[:, j])
                axs[j, i], caxs[j][i] = plot_spectr(
                    axs[j, i],
                    hist_,
                    cmap=cmap,
                    cscale="log",
                )
                axs[j, i].grid()

        out = (fig, axs, caxs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plasma_calc(b_xyz, t_i, t_e, n_i, n_e):
    """Computes plasma parameters including characteristic length and time
    scales.

    Parameters
    ----------
    b_xyz : xarray.DataArray
        Time series of the magnetic field [nT].
    t_i : xarray.DataArray
        Time series of the ions scalar temperature [eV].
    t_e : xarray.DataArray
        Time series of the electrons scalar temperature [eV].
    n_i : xarray.DataArray
        Time series of the ions number density [cm^{-3}].
    n_e : xarray.DataArray
        Time series of the electrons number density [cm^{-3}].

    Returns
    -------
    out : xarray.Dataset
        Dataset of the plasma parameters :
            * time : xarray.DataArray
                Time.
            * Wpe : xarray.DataArray
                Time series of the electron plasma frequency [rad.s^{-1}].
            * Fpe : xarray.DataArray
                Time series of the electron plasma frequency [Hz].
            * Wce : xarray.DataArray
                Time series of the electron cyclotron frequency [rad.s^{-1}].
            * Fce : xarray.DataArray
                Time series of the electron cyclotron frequency [Hz].
            * Wpp : xarray.DataArray
                Time series of the ion plasma frequency [rad.s^{-1}].
            * Fpp : xarray.DataArray
                Time series of the ion plasma frequency [Hz].
            * Fcp : xarray.DataArray
                Time series of the ion cyclotron frequency [Hz].
            * Fuh : xarray.DataArray
                Time series of the upper hybrid frequency [Hz].
            * Flh : xarray.DataArray
                Time series of the lower hybrid frequency [Hz].
            * Va : xarray.DataArray
                Time series of the Alfvèn velocity (ions) [m.s^{-1}].
            * Vae : xarray.DataArray
                Time series of the Alfvèn velocity (electrons) [m.s^{-1}].
            * Vte : xarray.DataArray
                Time series of the electron thermal velocity [m.s^{-1}].
            * Vtp : xarray.DataArray
                Time series of the electron thermal velocity [m.s^{-1}].
            * Vts : xarray.DataArray
                Time series of the sound speed [m.s^{-1}].
            * gamma_e : xarray.DataArray
                Time series of the electron Lorentz factor.
            * gamma_p : xarray.DataArray
                Time series of the electron Lorentz factor.
            * Le : xarray.DataArray
                Time series of the electron inertial length [m].
            * Li : xarray.DataArray
                Time series of the electron inertial length [m].
            * Ld : xarray.DataArray
                Time series of the Debye length [m].
            * Nd : xarray.DataArray
                Time series of the number of electrons in the Debye sphere.
            * Roe : xarray.DataArray
                Time series of the electron Larmor radius [m].
            * Rop : xarray.DataArray
                Time series of the ion Larmor radius [m].
            * Ros : xarray.DataArray
                Time series of the length associated to the sound speed [m].

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field, ion/electron temperature and number density

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> t_xyz_e = mms.get_data("Te_gse_fpi_fast_l2", tint, mms_id)
    >>> n_i = mms.get_data("Ni_fpi_fast_l2", tint, mms_id)
    >>> n_e = mms.get_data("Ne_fpi_fast_l2", tint, mms_id)

    Compute scalar temperature

    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")
    >>> t_xyzfac_e = mms.rotate_tensor(t_xyz_e, "fac", b_xyz, "pp")
    >>> t_i = pyrf.trace(t_xyzfac_i)
    >>> t_e = pyrf.trace(t_xyzfac_e)

    Compute plasma parameters

    >>> plasma_params = pyrf.plasma_calc(b_xyz, t_i, t_e, n_i, n_e)

    """

    # Get constants
    q_e = constants.elementary_charge
    cel = constants.speed_of_light
    mu0 = constants.mu_0
    ep0 = constants.epsilon_0
    m_p = constants.proton_mass
    m_e = constants.electron_mass
    mp_me = m_p / m_e

    # Resample all variables with respect to the magnetic field
    t_i = resample(t_i, b_xyz).data
    t_e = resample(t_e, b_xyz).data
    n_i = resample(n_i, b_xyz).data
    n_e = resample(n_e, b_xyz).data

    # Transform number density and magnetic field to SI units
    n_i, n_e = [1e6 * n_i, 1e6 * n_e]

    b_si = 1e-9 * np.linalg.norm(b_xyz, axis=1)

    w_pe = np.sqrt(n_e * q_e**2 / (m_e * ep0))  # rad/s
    w_ce = q_e * b_si / m_e  # rad/s
    w_pp = np.sqrt(n_i * q_e**2 / (m_p * ep0))

    v_a = b_si / np.sqrt(mu0 * n_i * m_p)

    v_ae = b_si / np.sqrt(mu0 * n_e * m_e)
    v_te = cel * np.sqrt(1 - 1 / (t_e * q_e / (m_e * cel**2) + 1) ** 2)
    v_tp = cel * np.sqrt(1 - 1 / (t_i * q_e / (m_p * cel**2) + 1) ** 2)
    # Sound speed formula (F. Chen, Springer 1984).
    v_ts = np.sqrt((t_e * q_e + 3 * t_i * q_e) / m_p)

    gamma_e = 1 / np.sqrt(1 - (v_te / cel) ** 2)
    gamma_p = 1 / np.sqrt(1 - (v_tp / cel) ** 2)

    l_e = cel / w_pe
    l_i = cel / w_pp
    # Debye length scale, sqrt(2) needed because of Vte definition
    l_d = v_te / (w_pe * np.sqrt(2))
    # number of e- in Debye sphere
    n_d = l_d * ep0 * m_e * v_te**2 / q_e**2

    f_pe = w_pe / (2 * np.pi)  # Hz
    f_ce = w_ce / (2 * np.pi)
    f_uh = np.sqrt(f_ce**2 + f_pe**2)
    f_pp = w_pp / (2 * np.pi)
    f_cp = f_ce / mp_me
    f_lh = np.sqrt(f_cp * f_ce / (1 + f_ce**2 / f_pe**2) + f_cp**2)

    rho_e = m_e * cel / (q_e * b_si) * np.sqrt(gamma_e**2 - 1)
    rho_p = m_p * cel / (q_e * b_si) * np.sqrt(gamma_p**2 - 1)
    rho_s = v_ts / (f_cp * 2 * np.pi)

    out = xr.Dataset(
        {
            "time": b_xyz.time.data,
            "w_pe": (["time"], w_pe),
            "w_ce": (["time"], w_ce),
            "w_pp": (["time"], w_pp),
            "v_a": (["time"], v_a),
            "v_ae": (["time"], v_ae),
            "v_te": (["time"], v_te),
            "v_tp": (["time"], v_tp),
            "v_ts": (["time"], v_ts),
            "gamma_e": (["time"], gamma_e),
            "gamma_p": (["time"], gamma_p),
            "l_e": (["time"], l_e),
            "l_i": (["time"], l_i),
            "l_d": (["time"], l_d),
            "n_d": (["time"], n_d),
            "f_pe": (["time"], f_pe),
            "f_ce": (["time"], f_ce),
            "f_uh": (["time"], f_uh),
            "f_pp": (["time"], f_pp),
            "f_cp": (["time"], f_cp),
            "f_lh": (["time"], f_lh),
            "rho_e": (["time"], rho_e),
            "rho_p": (["time"], rho_p),
            "rho_s": (["time"], rho_s),
        },
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants, interpolate

# Local imports
from pyrfu.pyrf import cart2sph, resample, sph2cart, time_clip

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["vdf_frame_transformation", "vdf_reduce"]


def _interp_skymap_sphe(vdf, energy, phi, theta, grid_sphe):
    r"""Interpolate the skymap distribution defined on the grid (`energy`,
    `phi`, `theta`) onto the spherical grid `grid_sphe`.

    Parameters
    ----------
    vdf : numpy.ndarray
        Values of the skymap distribution.
    energy : numpy.ndarray
        Energy level of skymap sampling.
    phi : numpy.ndarray
        Azimuthal angle of skymap sampling.
    theta : numpy.ndarray
        Elevation angle of skymap sampling.
    grid_sphe : numpy.ndarray
        Spherical velocity grid to interpolate on (3xlxmxn).

    Returns
    -------
    out_data : numpy.ndarray
        Values of the distribution interpolated onto `grdi_sphe`.

    Notes
    -----
    The values corresponding to energy levels below the instrument range are
    discarded.

    """

    phi_period = np.zeros(len(phi) + 2)
    phi_period[1:-1] = phi
    phi_period[0] = phi[-1] - 2 * 180.0
    phi_period[-1] = phi[0] + 2 * 180.0

    theta_period = np.zeros(len(theta) + 2)
    theta_period[1:-1] = theta
    theta_period[0] = theta[-1] - 180.0
    theta_period[-1] = theta[0] + 180.0

    vdf_period = np.zeros((len(energy), len(phi) + 2, len(theta) + 2))
    vdf_period[:, 1:-1, 1:-1] = vdf
    vdf_period[:, 1:-1, 0] = vdf[:, :, -1]
    vdf_period[:, 1:-1, -1] = vdf[:, :, 0]
    vdf_period[:, 0] = vdf_period[:, 1]
    vdf_period[:, -1] = vdf_period[:, -2]

    vdf_interp = interpolate.RegularGridInterpolator(
        (energy, phi_period, theta_period),
        vdf_period,
        method="linear",
        bounds_error=False,
        fill_value=None,
    )

    out_data = vdf_interp(grid_sphe)

    return out_data


def _interp_skymap_cart(vdf, energy, phi, theta, grid_cart):
    r"""Interpolate the skymap distribution defined on the grid (`energy`,
    `phi`, `theta`) onto the cartesian grid `grid_cart`.

    Parameters
    ----------
    vdf : numpy.ndarray
        Values of the skymap distribution.
    energy : numpy.ndarray
        Energy level of skymap sampling.
    phi : numpy.ndarray
        Azimuthal angle of skymap sampling.
    theta : numpy.ndarray
        Elevation angle of skymap sampling.
    grid_cart : numpy.ndarray
        Cartesian velocity grid to interpolate on (3xlxmxn).

    Returns
    -------
    out_data : numpy.ndarray
        Values of the distribution interpolated onto `grid_cart`.

    Notes
    -----
    The values corresponding to energy levels below the instrument range are
    discarded.

    See Also
    --------
    _inter_skymap_sphe.py
    """

    # Unpack cartesian grid
    v_x, v_y, v_z = grid_cart

    # Transform cartesian velocity grid to spherical velocity grid
    phi_grid, theta_grid, speed_grid = cart2sph(v_x, v_y, v_z)
    energy_grid = (
        0.5 * constants.proton_mass * speed_grid**2 / constants.elementary_charge
    )
    phi_grid = np.rad2deg(phi_grid) + 180.0
    theta_grid = np.rad2deg(theta_grid)

    grid_sphe = np.transpose(
        np.stack([energy_grid, phi_grid, theta_grid]),
        [1, 2, 3, 0],
    )

    # Interpolate the skymap distribution onto the spherical grid
    out_data = _interp_skymap_sphe(vdf, energy, phi, theta, grid_sphe)

    # Discard points with energy below the instrument energy range.
    v_min_2 = 2 * energy[0] * constants.electron_volt / constants.proton_mass
    out_data[v_x**2 + v_y**2 + v_z**2 < v_min_2] = np.nan

    return out_data


def vdf_frame_transformation(vdf, v_gse):
    r"""Move the skymap into the desired frame associated with the bulk
    velocity `v_gse`.

    Parameters
    ----------
    vdf : xarray.Dataset
        Skymap distribution in the initial frame.
    v_gse : xarray.DataArray
        Time series of the bulk velocity to shift.

    Returns
    -------
    out : xarray.Dataset
        Skymap distribution into the new frame.

    Notes
    -----
    The new skymap grid is identical to the original one. The bulk velocity
    must be in the same coordinates system as the skymap (i.e spacecraft for
    FPI and GSE for EIS)

    See Also
    --------
    _interp_skymap_cart.py, _interp_skymap_sphe.py

    """

    v_gse = resample(v_gse, vdf.time)
    theta = vdf.theta.data

    out_data = np.zeros_like(vdf.data.data)

    for i in range(len(vdf.time.data)):
        vdf_data = vdf.data.data[i, :]
        energy = vdf.energy.data[i, :]
        phi = vdf.phi.data[i, :]

        phi_mat, en_mat, theta_mat = np.meshgrid(phi, energy, theta)
        v_mat = np.sqrt(
            2 * en_mat * constants.electron_volt / constants.proton_mass,
        )
        v_xyz = sph2cart(np.deg2rad(phi_mat), np.deg2rad(theta_mat), v_mat)

        grid_cart = np.stack(
            [
                v_xyz[0] - v_gse.data[i, 0, None, None, None],
                v_xyz[1] - v_gse.data[i, 1, None, None, None],
                v_xyz[2] - v_gse.data[i, 2, None, None, None],
            ],
        )

        out_data[i, ...] = _interp_skymap_cart(
            vdf_data,
            energy,
            phi,
            theta,
            grid_cart,
        )

    out = vdf.copy()
    out.data.data = out_data

    return out


def vdf_reduce(
    vdf,
    tint,
    dim,
    x_vec,
    z_vec: list = None,
    v_int: list = None,
    n_vpt: int = 100,
):
    r"""Interpolate the skymap distribution onto the velocity grid defined
    by the velocity interval `v_int` along the axes `x_vec` and `z_vec`,
    and reduce (integrate) it along 1 (if `dim` is "2d") or 2 (if `dim` is
    "1d").

    Parameters
    ----------
    vdf : xarray.Dataset
        Skymap distribution to reduce.
    tint : list of strs
        Time interval over which the time series of the skymap distribution
        is averaged.
    dim : {"1d", "2d"}
        Dimension of the output reduced distribution.
    x_vec : array_like
        X axis. For the "1d" case, it is the axis on which the skymap is
        plotted. For the "2d" case, it is the first of the two axes on which
        the skymap is plotted.
    z_vec : array_like, Optional
        Axis along which the skymap is integrated. Needed only for the "2d"
        case.
    v_int : array_like, Optional
        Velocity interval.
    n_vpt : int, Optional
        Number of points along the plot direction(s).

    Returns
    -------
    out : xarray.DataArray
        Reduced distribution.

    """

    if v_int is None:
        v_int = [-1e6, 1e6]
    if z_vec is None:
        z_vec = [0, 0, 1]

    x_vec = x_vec / np.linalg.norm(x_vec, keepdims=True)
    y_vec = np.cross(z_vec, x_vec)
    y_vec = y_vec / np.linalg.norm(y_vec, keepdims=True)
    z_vec = np.cross(x_vec, y_vec)
    z_vec = z_vec / np.linalg.norm(z_vec, keepdims=True)

    v_x, v_y, v_z = [np.linspace(v_int[0], v_int[1], n_vpt) for _ in range(3)]

    m_vec = np.transpose([x_vec, y_vec, z_vec])
    v_x, v_y, v_z = np.matmul(np.linalg.inv(m_vec), np.array([v_x, v_y, v_z]))
    v_x_mat, v_y_mat, v_z_mat = np.meshgrid(v_x, v_y, v_z)
    grid_cart = np.stack([v_x_mat, v_y_mat, v_z_mat])

    vdf = time_clip(vdf, tint)

    vdf_data = np.mean(vdf.data.data, axis=0)
    energy = np.mean(np.atleast_2d(vdf.energy.data), axis=0)
    phi = np.mean(np.atleast_2d(vdf.phi.data), axis=0)
    theta = vdf.theta.data

    interp_vdf = _interp_skymap_cart(vdf_data, energy, phi, theta, grid_cart)

    if dim.lower() == "2d":
        dv_ = np.abs(np.diff(v_z)[0])
        red_vdf = np.nansum(interp_vdf, axis=-1) * dv_
        out = xr.DataArray(
            red_vdf,
            coords=[v_x / 1e6, v_y / 1e6],
            dims=["vx", "vy"],
        )
    elif dim.lower() == "1d":
        dv_ = np.abs(np.diff(v_y)[0] * np.diff(v_z)[0])
        red_vdf = np.nansum(np.sum(interp_vdf, axis=-1), axis=-1) * dv_
        out = xr.DataArray(red_vdf, coords=[v_x / 1e6], dims=["vx"])

    else:
        raise ValueError

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def timevec2iso8601(time):
    r"""Convert time vector into ISO 8601 format YYYY-MM-DDThh:mm:ss.mmmuuunnn.

    Parameters
    ----------
    time : ndarray
        Time vector

    Returns
    -------
    time_iso8601 : ndarray
        Time in ISO 8601 format.

    See Also
    --------
    pyrfu.pyrf.iso86012timevec

    """

    time = np.atleast_2d(np.array(time))
    time = np.hstack([time, np.zeros((len(time), 9 - time.shape[1]))])

    time_iso8601 = []

    for t_ in time.astype(np.int64):
        ye_mo_da_ = f"{t_[0]:04}-{t_[1]:02}-{t_[2]:02}"  # YYYY-MM-DD
        ho_mi_se_ = f"{t_[3]:02}:{t_[4]:02}:{t_[5]:02}"  # hh:mm:ss
        ms_us_ns_ = f"{t_[6]:03}{t_[7]:03}{t_[8]:03}"  # mmmuuunnn

        # Time as ISO 8601 string 'YYYY-MM-DDThh:mm:ss.mmmuuunnn'
        time_iso8601.append(f"{ye_mo_da_}T{ho_mi_se_}.{ms_us_ns_}")

    time_iso8601 = np.array(time_iso8601)

    return time_iso8601



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import fnmatch
import glob
import json

# Built-in imports
import os
import re

# 3rd party imports
import numpy as np
import pandas as pd

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.iso86012datetime import iso86012datetime
from ..pyrf.iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.11"
__status__ = "Prototype"


def list_files_ancillary(tint, mms_id, product, data_path: str = ""):
    r"""Find available ancillary files in the data directories for the target product
    type.

    Parameters
    ----------
    tint : list of str
        Time interval
    mms_id : str or int
        Spacecraft index
    product : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu.mms.mms_config.py`

    Returns
    -------
    file_names : list
        Ancillary files in interval.

    """

    # Check path
    if not data_path:
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MMS configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    # Make sure that the data path exists
    assert os.path.exists(data_path), f"{data_path} doesn't exist!!"

    if isinstance(mms_id, int):
        mms_id = str(mms_id)

    # Check time interval type
    if isinstance(tint, (np.ndarray, list)):
        if isinstance(tint[0], np.datetime64):
            tint = datetime642iso8601(tint)
            tint = iso86012datetime(tint)
        elif isinstance(tint[0], str):
            tint = iso86012datetime64(
                np.array(tint),
            )  # to make sure it is ISO8601 ok!
            tint = datetime642iso8601(tint)
            tint = iso86012datetime(tint)
        elif isinstance(tint[0], datetime.datetime):
            pass
        else:
            raise TypeError("Values must be in datetime, datetime64, or str!!")
    else:
        raise TypeError("tint must be a DataArray or array_like!!")

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    dir_pattern = os.sep.join([data_path, "ancillary", f"mms{mms_id}", product])
    file_pattern = "_".join([f"MMS{mms_id}", product.upper(), "???????_???????.V??"])

    files_in_tint = []
    out_files = []

    files = glob.glob(os.sep.join([dir_pattern, file_pattern]))

    # find the files within the time interval
    fname_fmt = f"MMS{mms_id}_{product.upper()}_([0-9]{{7}})_([0-9]{{7}}).V[0-9]{{2}}"

    if os.name == "nt":
        full_path = os.sep.join([re.escape(dir_pattern) + os.sep, fname_fmt])
    else:
        full_path = os.sep.join([re.escape(dir_pattern), fname_fmt])

    file_regex = re.compile(full_path)

    for file in files:
        time_match = file_regex.match(file)
        if time_match is not None:
            start_time = pd.to_datetime(time_match.group(1), format="%Y%j")
            end_time = pd.to_datetime(time_match.group(2), format="%Y%j")
            if start_time < tint[1] and end_time >= tint[0]:
                files_in_tint.append(file)

    # ensure only the latest version of each file is loaded
    for file in files_in_tint:
        this_file = file[0:-3] + "V??"
        versions = fnmatch.filter(files_in_tint, this_file)
        if len(versions) > 1:
            # only grab the latest version
            out_files.append(sorted(versions)[-1])
        else:
            out_files.append(versions[0])

    file_names = list(set(out_files))
    file_names.sort()

    return file_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect
import logging

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.avg_4sc import avg_4sc
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip
from ..pyrf.wavelet import wavelet

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def fk_power_spectrum_4sc(
    e,
    r,
    b,
    tints,
    cav: int = 8,
    num_k: int = 500,
    num_f: int = 200,
    df: float = None,
    w_width: int = 1,
    f_range: list = None,
):
    r"""Calculates the frequency-wave number power spectrum using the four
    MMS spacecraft. Uses a generalization of mms.fk_powerspectrum. Wavelet
    based cross-spectral analysis is used to calculate the phase difference
    each spacecraft pair and determine 3D wave vector. A generalization of
    the method used in mms.fk_powerspectrum to four point measurements.

    Parameters
    ----------
    e : list of xarray.DataArray
        Fields to apply 4SC cross-spectral analysis to. e.g., e or b fields
        (if multiple components only the first is used).
    r : list of xarray.DataArray
        Positions of the four spacecraft.
    b : list of xarray.DataArray
        background magnetic field in the same coordinates as r. Used to
        determine the parallel and perpendicular wave numbers using 4SC
        average.
    tints : list of str
        Time interval over which the power spectrum is calculated. To avoid
        boundary effects use a longer time interval for e and b.
    cav : int, Optional
        Number of points in time series used to estimate phase.
        Default ``cav`` = 8.
    num_k : int, Optional
        Number of wave numbers used in spectrogram. Default ``num_k`` = 500.
    df : float, Optional
        Linear spacing of frequencies. Default ``df`` = None (log spacing).
    num_f : int, Optional
        Number of frequencies used in spectrogram. Default ``num_f`` = 200.
    w_width : float, Optional
        Multiplier for Morlet wavelet width. Default ``w_width`` = 1.
    f_range : list of float, Optional
        Frequency range for k-k plots. [minf maxf].

    Returns
    -------
    out : xarray.Dataset
        Dataset of array of powers as a function of frequency and
        wavenumber. Power is normalized to the maximum value.

    Notes
    -----
    Wavelength must be larger than twice the spacecraft separations,
    otherwise spatial aliasing will occur.

    Examples
    --------
    >>> from pyrfu.mms import get_data, fk_power_spectrum_4sc
    >>> from pyrfu.pyrf import extend_tint, convert_fac

    Load data

    >>> tint = ["2015-10-16T13:05:24.00", "2015-10-16T13:05:50.000"]
    >>> ic = range(1, 5)
    >>> b_fgm = [get_data("b_gse_fgm_brst_l2", tint, i) for i in ic]
    >>> b_scm = [get_data("b_gse_scm_brst_l2", tint, i) for i in ic]

    Load spacecraft position

    >>> tint_long = extend_tint(tint, [-60, 60])
    >>> r_gse_mms = [get_data("r_gse", tint_long, i) for i in range(1, 5)]

    Convert magnetic field fluctuations to field aligned coordinates

    >>> b_fac = [convert_fac(b_s, b_f) for b_s, b_f in zip(b_scm, b_fgm)]
    >>> b_par = [b_s[:, 0] for b_s in b_fac]

    Compute dispersion relation

    >>> tint = ["2015-10-16T13:05:26.500", "2015-10-16T13:05:27.000"]
    >>> pwer = fk_power_spectrum_4sc(b_par, r_gse, b_fgm, tint, 4, 500, 2,
    ... 10, 2)

    """

    ic = np.arange(1, 5)

    e = [resample(e[i - 1], e[0]) for i in ic]
    r = [resample(r[i - 1], e[0]) for i in ic]
    b = [resample(b[i - 1], e[0]) for i in ic]

    b_avg = avg_4sc(b)

    times = e[0].time
    use_linear = df is not None

    # idx = time_clip(e[0].time, list(tints))

    # If odd, remove last data point (as is done in irf_wavelet)
    # if len(idx) % 2:
    #     idx = idx[:-1]

    if use_linear:
        cwt_options = {
            "linear": df,
            "return_power": False,
            "wavelet_width": 5.36 * w_width,
            "cut_edge": False,
        }
    else:
        cwt_options = {
            "nf": num_f,
            "return_power": False,
            "wavelet_width": 5.36 * w_width,
            "cut_edge": False,
        }

    w = [wavelet(e[i], **cwt_options) for i in range(4)]

    num_f = len(w[0].frequency)

    times = time_clip(times, tints)
    nt = len(times)

    w = [time_clip(w[i], tints) for i in range(4)]

    fk_power = 0
    for i in range(4):
        fk_power += (w[i].data * np.conj(w[i].data) / 4).astype(np.float64)

    n = int(np.floor(nt / cav) - 1)
    pos_av = cav / 2 + np.arange(n + 1) * cav
    av_times = times[pos_av.astype(np.int64)]

    b_avg = resample(b_avg, av_times)

    r = [resample(r[i], av_times) for i in range(4)]

    cx12, cx13, cx14 = [np.zeros((n + 1, num_f), dtype="complex128") for _ in range(3)]
    cx23, cx24, cx34 = [np.zeros((n + 1, num_f), dtype="complex128") for _ in range(3)]

    power_avg = np.zeros((n + 1, num_f), dtype="complex128")

    for m, pos_avm in enumerate(pos_av):
        lb, ub = [int(pos_avm - cav / 2), int(pos_avm + cav / 2)]

        cx12[m, :] = np.nanmean(
            w[0].data[lb:ub, :] * np.conj(w[1].data[lb:ub, :]),
            axis=0,
        )
        cx13[m, :] = np.nanmean(
            w[0].data[lb:ub, :] * np.conj(w[2].data[lb:ub, :]),
            axis=0,
        )
        cx14[m, :] = np.nanmean(
            w[0].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]),
            axis=0,
        )
        cx23[m, :] = np.nanmean(
            w[1].data[lb:ub, :] * np.conj(w[2].data[lb:ub, :]),
            axis=0,
        )
        cx24[m, :] = np.nanmean(
            w[1].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]),
            axis=0,
        )
        cx34[m, :] = np.nanmean(
            w[2].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]),
            axis=0,
        )

        power_avg[m, :] = np.nanmean(fk_power[lb:ub, :], axis=0)

    # Compute phase differences between each spacecraft pair
    th12 = np.arctan2(np.imag(cx12), np.real(cx12))
    th13 = np.arctan2(np.imag(cx13), np.real(cx13))
    th14 = np.arctan2(np.imag(cx14), np.real(cx14))
    th23 = np.arctan2(np.imag(cx23), np.real(cx23))
    th24 = np.arctan2(np.imag(cx24), np.real(cx24))
    th34 = np.arctan2(np.imag(cx34), np.real(cx34))

    w_mat = 2 * np.pi * np.tile(w[0].frequency.data, (n + 1, 1))

    # Convert phase difference to time delay
    dt12, dt13, dt14, dt23, dt24, dt34 = [
        th / w_mat for th in [th12, th13, th14, th23, th24, th34]
    ]

    # Weighted averaged time delay using all spacecraft pairs
    dt2 = (
        0.5 * dt12
        + 0.2 * (dt13 - dt23)
        + 0.2 * (dt14 - dt24)
        + 0.1 * (dt14 - dt34 - dt23)
    )
    dt3 = (
        0.5 * dt13
        + 0.2 * (dt12 + dt23)
        + 0.2 * (dt14 - dt34)
        + 0.1 * (dt12 + dt24 - dt34)
    )
    dt4 = (
        0.5 * dt14
        + 0.2 * (dt12 + dt24)
        + 0.2 * (dt13 + dt34)
        + 0.1 * (dt12 + dt23 + dt34)
    )

    # Compute phase speeds
    r = [r[i].data for i in range(4)]

    k_x, k_y, k_z = [np.zeros((n + 1, num_f)) for _ in range(3)]

    for ii in range(n + 1):
        dr = np.array(
            [
                r[1][ii, :] - r[0][ii, :],
                r[2][ii, :] - r[0][ii, :],
                r[3][ii, :] - r[0][ii, :],
            ],
        )
        for jj in range(num_f):
            m = np.linalg.solve(dr, [dt2[ii, jj], dt3[ii, jj], dt4[ii, jj]])
            k_x[ii, jj] = 2 * np.pi * w[0].frequency[jj].data * m[0]
            k_y[ii, jj] = 2 * np.pi * w[0].frequency[jj].data * m[1]
            k_z[ii, jj] = 2 * np.pi * w[0].frequency[jj].data * m[2]

    k_x, k_y, k_z = [k / 1e3 for k in [k_x, k_y, k_z]]

    k_mag = np.linalg.norm(np.array([k_x, k_y, k_z]), axis=0)

    b_avg_x_mat = np.tile(b_avg.data[:, 0], (num_f, 1)).T
    b_avg_y_mat = np.tile(b_avg.data[:, 1], (num_f, 1)).T
    b_avg_z_mat = np.tile(b_avg.data[:, 2], (num_f, 1)).T

    b_avg_abs = np.linalg.norm(b_avg, axis=1)
    b_avg_abs_mat = np.tile(b_avg_abs, (num_f, 1)).T

    k_par = (k_x * b_avg_x_mat + k_y * b_avg_y_mat + k_z * b_avg_z_mat) / b_avg_abs_mat
    k_perp = np.sqrt(k_mag**2 - k_par**2)

    k_max = np.max(k_mag) * 1.1
    k_min = -k_max
    k_vec = np.linspace(-k_max, k_max, num_k)
    k_mag_vec = np.linspace(0, k_max, num_k)

    dk_mag = k_max / num_k
    dk = 2 * k_max / num_k

    # Sort power into frequency and wave vector
    logging.info("Computing power versus kx,f; ky,f, kz,f")
    power_k_x_f, power_k_y_f, power_k_z_f = [np.zeros((num_f, num_k)) for _ in range(3)]
    power_k_mag_f = np.zeros((num_f, num_k))

    for nn in range(num_f):
        k_x_number = np.floor((k_x[:, nn] - k_min) / dk).astype(np.int64)
        k_y_number = np.floor((k_y[:, nn] - k_min) / dk).astype(np.int64)
        k_z_number = np.floor((k_z[:, nn] - k_min) / dk).astype(np.int64)
        k_number = np.floor((k_mag[:, nn]) / dk_mag).astype(np.int64)

        power_k_x_f[nn, k_x_number] += np.real(power_avg[:, nn])
        power_k_y_f[nn, k_y_number] += np.real(power_avg[:, nn])
        power_k_z_f[nn, k_z_number] += np.real(power_avg[:, nn])

        power_k_mag_f[nn, k_number] += np.real(power_avg[:, nn])

    power_k_x_f /= np.max(power_k_x_f)
    power_k_y_f /= np.max(power_k_y_f)
    power_k_z_f /= np.max(power_k_z_f)
    power_k_mag_f /= np.max(power_k_mag_f)

    frequencies = w[0].frequency.data
    idx_f = np.arange(num_f)

    if f_range is not None:
        idx_min_freq = bisect.bisect_left(frequencies, np.min(f_range))
        idx_max_freq = bisect.bisect_left(frequencies, np.max(f_range))
        idx_f = idx_f[idx_min_freq:idx_max_freq]

    logging.info("Computing power versus kx,ky; kx,kz; ky,kz")
    power_k_x_k_y = np.zeros((num_k, num_k))
    power_k_x_k_z = np.zeros((num_k, num_k))
    power_k_y_k_z = np.zeros((num_k, num_k))
    power_k_perp_k_par = np.zeros((num_k, num_k))

    for nn in idx_f:
        k_x_number = np.floor((k_x[:, nn] - k_min) / dk).astype(np.int64)
        k_y_number = np.floor((k_y[:, nn] - k_min) / dk).astype(np.int64)
        k_z_number = np.floor((k_z[:, nn] - k_min) / dk).astype(np.int64)

        k_par_number = np.floor((k_par[:, nn] - k_min) / dk).astype(np.int64)
        k_perp_number = np.floor((k_perp[:, nn]) / dk_mag).astype(np.int64)

        power_k_x_k_y[k_y_number, k_x_number] += np.real(power_avg[:, nn])
        power_k_x_k_z[k_z_number, k_x_number] += np.real(power_avg[:, nn])
        power_k_y_k_z[k_z_number, k_y_number] += np.real(power_avg[:, nn])

        power_k_perp_k_par[k_par_number, k_perp_number] += np.real(
            power_avg[:, nn],
        )

    power_k_x_k_y /= np.max(power_k_x_k_y)
    power_k_x_k_z /= np.max(power_k_x_k_z)
    power_k_y_k_z /= np.max(power_k_y_k_z)
    power_k_perp_k_par /= np.max(power_k_perp_k_par)

    out_dict = {
        "k_x_f": (["k_x", "f"], power_k_x_f.T),
        "k_y_f": (["k_x", "f"], power_k_y_f.T),
        "k_z_f": (["k_x", "f"], power_k_z_f.T),
        "k_mag_f": (["k_mag", "f"], power_k_mag_f.T),
        "k_x_k_y": (["k_x", "k_y"], power_k_x_k_y.T),
        "k_x_k_z": (["kx", "kz"], power_k_x_k_z.T),
        "k_y_k_z": (["k_y", "k_z"], power_k_y_k_z.T),
        "k_perp_k_par": (["k_perp", "k_par"], power_k_perp_k_par.T),
        "k_x": k_vec,
        "k_y": k_vec,
        "k_z": k_vec,
        "k_mag": k_mag_vec,
        "k_perp": k_mag_vec,
        "k_par": k_vec,
        "f": frequencies,
    }

    out = xr.Dataset(out_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
from matplotlib import dates

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def span_tint(axs, tint, ymin: float = 0, ymax: float = 1, **kwargs):
    r"""Add a vertical span (rectangle) across the time Axes.

    The rectangle spans from tint[0] to tint[1] horizontally, and, by default,
    the whole y-axis vertically. The y-span can be set using ymin (default: 0)
    and ymax (default: 1) which are in axis units; e.g. ymin = 0.5 always
    refers to the middle of the y-axis regardless of the limits set by
    set_ylim

    Parameters
    ----------
    axs : list of matplotlib.pyplot.subplotsaxes
        Axes to span.
    tint : list of str
        Time interval to span
    ymin : float
        Lower y-coordinate of the span, in y-axis units (0-1). Default ymin=0.
    ymax : float, Optional
        Upper y-coordinate of the span, in y-axis units (0-1). Default ymax=1.

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments control the Polygon properties.

    Returns
    -------
    axs : list
        List of matplotlib.axes._subplots.AxesSubplot spanned..

    """

    for axis in axs:
        t_start, t_stop = [
            dates.datestr2num(tint[0]),
            dates.datestr2num(tint[1]),
        ]
        axis.axvspan(t_start, t_stop, ymin, ymax, **kwargs)

    return axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging
import os

# 3rd party imports
import numba
import numpy as np
import xarray as xr
from scipy import fft

# Local imports
from .calc_fs import calc_fs

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


@numba.jit(nopython=True, fastmath=True)
def _ww(s_ww, scales_mat, sigma, frequencies_mat, f_nyq):
    # TODO : use nested for loop and math instead of numpy and test speed!!
    w_w = s_ww * np.exp(
        -sigma * sigma * ((scales_mat * frequencies_mat - f_nyq) ** 2) / 2,
    )
    w_w = w_w * np.sqrt(1)
    return w_w


@numba.jit(nopython=True, parallel=True, fastmath=True)
def _power_r(power, new_freq_mat):
    power2 = np.absolute((2 * np.pi) * np.conj(power) * power / new_freq_mat)
    return power2


@numba.jit(nopython=True, parallel=True, fastmath=True)
def _power_c(power, new_freq_mat):
    power2 = np.sqrt(np.absolute((2 * np.pi) / new_freq_mat)) * power
    return power2


def wavelet(inp, **kwargs):
    """Computes wavelet spectrogram based on fast FFT algorithm.
    Parameters
    ----------
    inp : xarray.DataArray
        Input quantity.
    **kwargs : dict
        Hash table of keyword arguments with :
            * fs : int or float
                Sampling frequency of the input time series.
            * f : list or ndarray
                Vector [f_min f_max], calculate spectra between frequencies
                f_min and f_max.
            * nf : int or float
                Number of frequency bins.
            * wavelet_width : int or float
                Width of the Morlet wavelet. Default 5.36.
            * linear : float
                Linear spacing between frequencies of df.
            * return_power : bool
                Set to True to return the power, False for complex wavelet
                transform. Default True.
            * cut_edge : bool
                Set to True to set points affected by edge effects to NaN,
                False to keep edge affect points. Default True
    Returns
    -------
    out : xarray.DataArray or xarray.Dataset
        Wavelet transform of the input.
    """

    # Unpack time and data
    data = inp.data

    f_s = kwargs.get("fs", calc_fs(inp))
    n_freqs = kwargs.get("nf", 200)
    wavelet_width = kwargs.get("wavelet_width", 5.36)
    cut_edge = kwargs.get("cut_edge", True)
    return_power = kwargs.get("return_power", True)

    if kwargs.get("linear"):
        linear_df = True
        if isinstance(kwargs["linear"], bool) and kwargs["linear"]:
            delta_f = 100
            logging.warning("Unknown input for linear delta_f set to 100")
        elif isinstance(kwargs["linear"], (int, float)):
            delta_f = kwargs["linear"]
        else:
            raise TypeError("linear keyword argument must be bool, float or int")
    else:
        delta_f = 100
        linear_df = False

    scale_min, scale_max = [0.01, 2]

    f_min, f_max = kwargs.get(
        "f",
        [0.5 * f_s / 10**scale_max, 0.5 * f_s / 10**scale_min],
    )

    f_nyq, scale_number, sigma = [f_s / 2, n_freqs, wavelet_width / (f_s / 2)]

    if linear_df:
        scale_number = np.floor(f_nyq / delta_f).astype(np.int64)

        f_min, f_max = [delta_f, scale_number * delta_f]

        scales = f_nyq / (np.linspace(f_max, f_min, scale_number))
    else:
        scale_min = np.log10(0.5 * f_s / f_max)
        scale_max = np.log10(0.5 * f_s / f_min)
        scales = np.logspace(scale_min, scale_max, scale_number)

    # Remove the last sample if the total number of samples is odd
    if len(data) / 2 != np.floor(len(data) / 2):
        time = inp.time.data[:-1]
        data = data[:-1, ...]
    else:
        time = inp.time.data

    # Check for NaNs
    scales[np.isnan(scales)] = 0

    # Find the frequencies for an FFT of all data
    freq = f_s * 0.5 * np.arange(1, 1 + len(data) / 2) / (len(data) / 2)

    # The frequencies corresponding to FFT
    frequencies = np.hstack([0, freq, -np.flip(freq[:-1])])

    # Get the correct frequencies for the wavelet transform
    new_freq = f_nyq / scales

    if len(inp.shape) == 1:
        out_dict, power2 = [None, np.zeros((len(inp.data), n_freqs))]
    elif len(inp.shape) == 2:
        out_dict, power2 = [{}, None]
    else:
        raise TypeError("Invalid shape of the inp")

    new_freq_mat, _ = np.meshgrid(new_freq, frequencies, sparse=True)

    _, frequencies_mat = np.meshgrid(scales, frequencies, sparse=True)

    # if scalar add virtual axis
    if len(inp.shape) == 1:
        data = data[:, np.newaxis]

    # go through all the data columns
    for i in range(data.shape[1]):
        # Make the FFT of all data
        data_col = data[:, i]

        # Wavelet transform of the data
        # Forward FFT
        s_w = fft.fft(data_col, workers=os.cpu_count())

        scales_mat, s_w_mat = np.meshgrid(scales, s_w, sparse=True)

        # Calculate the FFT of the wavelet transform
        w_w = _ww(s_w_mat, scales_mat, sigma, frequencies_mat, f_nyq)

        # Backward FFT
        power = fft.ifft(w_w, axis=0, workers=os.cpu_count())

        # Calculate the power spectrum
        if return_power:
            power2 = _power_r(power, np.tile(new_freq_mat, (len(power), 1)))
        else:
            power2 = _power_c(power, np.tile(new_freq_mat, (len(power), 1)))

        if cut_edge:
            censure = np.floor(2 * scales).astype(int)

            for j in range(scale_number):
                power2[: censure[j], j] = np.nan

                power2[len(data_col) - censure[j] : len(data_col), j] = np.nan

        if len(inp.shape) == 2:
            out_dict[inp.comp.data[i]] = (
                ["time", "frequency"],
                np.fliplr(power2),
            )

    if len(inp.shape) == 1:
        out = xr.DataArray(
            np.fliplr(power2),
            coords=[time, np.flip(new_freq)],
            dims=["time", "frequency"],
        )
    else:
        out = xr.Dataset(
            out_dict,
            coords={"time": time, "frequency": np.flip(new_freq)},
        )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def dynamic_press(n_s, v_xyz, specie: str = "ions"):
    r"""Computes dynamic pressure.

    Parameters
    ----------
    n_s : xarray.DataArray
        Time series of the number density of the specie.
    v_xyz : xarray.DataArray
        Time series of the bulk velocity of the specie.
    specie : {"ions", "electrons"}, Optional
        Specie. Default "ions".

    Returns
    -------
    p_dyn : xarray.DataArray
        Time series of the dynamic pressure of the specie.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load ion bulk velocity and remove spintone

    >>> v_xyz_i = mms.get_data("vi_gse_fpi_fast_l2", tint, mms_id)
    >>> st_xyz_i = mms.get_data("sti_gse_fpi_fast_l2", tint, mms_id)
    >>> v_xyz_i = v_xyz_i - st_xyz_i

    Ion number density

    >>> n_i = mms.get_data("ni_fpi_fast_l2", tint, mms_id)

    Compute dynamic pressure

    >>> p = pyrf.dynamic_press(n_i, v_xyz_i, specie="ions")

    """

    # Check input
    assert isinstance(n_s, xr.DataArray), "n_s must be a xarray.DataArray"
    assert isinstance(v_xyz, xr.DataArray), "v_xyz must be a xarray.DataArray"
    assert isinstance(specie, str), "specie must be a str"
    assert specie.lower() in ["ions", "electrons"], "specie must be ions or electrons"

    # Check n_s and v_xyz shapes
    assert n_s.ndim == 1, "n_s must be a scalar"
    assert v_xyz.ndim == 2 and v_xyz.shape[1] == 3, "v_xyz must be a vector"

    if specie.lower() == "ions":
        mass = constants.proton_mass
    else:
        mass = constants.electron_mass

    p_dyn = n_s * mass * np.linalg.norm(v_xyz, axis=1) ** 2

    return p_dyn



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os
import random
import unittest

# 3rd party imports
import numpy as np
from ddt import data, ddt, unpack

from pyrfu import solo

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.4"
__status__ = "Prototype"


class DbInitTestCase(unittest.TestCase):
    def test_db_init_inpput(self):
        with self.assertRaises(AssertionError):
            solo.db_init("/Volumes/solo/remote/data")

    def test_db_init_output(self):
        self.assertIsNone(solo.db_init(os.getcwd()))


@ddt
class ReadLFRDensityTestCase(unittest.TestCase):
    @data(
        ([], ".", False),
        ([np.datetime64("2023-01-01T00:00:00"), "2023-01-01T00:10:00"], ".", False),
        (["2023-01-01T00:00:00", np.datetime64("2023-01-01T00:10:00")], ".", False),
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], "/bazinga", False),
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], ".", "i am groot"),
    )
    @unpack
    def test_read_lfr_density_input(self, tint, data_path, tree):
        with self.assertRaises(AssertionError):
            solo.read_lfr_density(tint, data_path, tree)

    def test_read_lfr_density_output(self):
        tint = ["2023-01-01T00:00:00", "2023-01-01T00:10:00"]
        self.assertIsNone(solo.read_lfr_density(tint))


@ddt
class ReadTNRTestCase(unittest.TestCase):
    @data(
        ([], 1, "."),
        ([np.datetime64("2023-01-01T00:00:00"), "2023-01-01T00:10:00"], 1, "."),
        (["2023-01-01T00:00:00", np.datetime64("2023-01-01T00:10:00")], 1, "."),
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], random.random(), "."),
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], 1, "/bazinga"),
    )
    @unpack
    def test_read_tnr_input(self, tint, sensor, data_path):
        with self.assertRaises(AssertionError):
            solo.read_tnr(tint, sensor, data_path)

    @data(
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], 1, ""),
        (["2023-01-01T00:00:00", "2023-01-01T00:10:00"], 2, ""),
    )
    @unpack
    def test_read_tnr_output(self, tint, sensor, data_path):
        self.assertIsNone(solo.read_tnr(tint, sensor, data_path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

# Local imports
from ..pyrf.resample import resample
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

q_e = constants.elementary_charge


def calculate_epsilon(vdf, model_vdf, n_s, sc_pot, **kwargs):
    r"""Calculates epsilon parameter using model distribution.

    Parameters
    ----------
    vdf : xarray.Dataset
        Observed particle distribution (skymap).
    model_vdf : xarray.Dataset
        Model particle distribution (skymap).
    n_s : xarray.DataArray
        Time series of the number density.
    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    epsilon : xarray.DataArray
        Time series of the epsilon parameter.

    Other Parameters
    ----------------
    en_channels : array_like
        Set energy channels to integrate over [min max]; min and max between
        must be between 1 and 32.

    Examples
    --------
    >>> from pyrfu import mms
    >>> options = {"en_channel": [4, 32]}
    >>> eps = mms.calculate_epsilon(vdf, model_vdf, n_s, sc_pot, **options)

    """

    # Resample sc_pot
    sc_pot = resample(sc_pot, n_s)

    vdf_data = vdf.data.data.copy() * 1e12
    model_vdf_data = model_vdf.data.data.copy() * 1e-18

    energy = vdf.energy.data.copy()
    phi = vdf.phi.data.copy()
    theta = vdf.theta.data.copy()

    vdf_diff = np.abs(vdf_data - model_vdf_data)

    if vdf.attrs["species"][0].lower() == "e":
        m_s = constants.electron_mass
    elif vdf.attrs["species"][0].lower() == "i":
        sc_pot.data *= -1
        m_s = constants.proton_mass
    else:
        raise ValueError("Invalid specie")

    if np.abs(np.median(np.diff(vdf.time.data - n_s.time.data))) > 0:
        raise ValueError("vdf and moments have different times.")

    # Default energy channels used to compute epsilon.
    energy_range = kwargs.get("en_channels", [0, vdf.energy.shape[1]])
    int_energies = np.arange(energy_range[0], energy_range[1])

    flag_same_e = np.sum(np.abs(vdf.attrs["energy0"] - vdf.attrs["energy1"])) < 1e-4

    # Calculate angle differences
    delta_phi = np.deg2rad(np.median(np.diff(phi[0, :])))
    delta_theta = np.deg2rad(np.median(np.diff(theta)))

    delta_ang = delta_phi * delta_theta

    phi_tr = phi.copy()
    theta_tr = np.tile(theta, (len(vdf.time.data), 1))

    energy_minus = vdf.attrs["delta_energy_minus"]
    energy_plus = vdf.attrs["delta_energy_plus"]

    # Calculate speed widths associated with each energy channel.
    energy_scpot = np.transpose(np.tile(sc_pot.data, (energy.shape[1], 1)))
    energy_corr = energy - np.transpose(
        np.tile(sc_pot.data, (energy.shape[1], 1)),
    )
    velocity = np.real(np.sqrt(2 * q_e * energy_corr / m_s))

    if flag_same_e:
        energy_upper = energy + energy_plus
        energy_lower = energy - energy_minus
        v_upper = np.sqrt(2 * q_e * (energy_upper - energy_scpot) / m_s)
        v_lower = np.sqrt(2 * q_e * (energy_lower - energy_scpot) / m_s)

    else:
        energy_upper = energy + energy_plus
        energy_lower = energy - energy_minus
        v_upper = np.sqrt(2 * q_e * (energy_upper - energy_scpot) / m_s)
        v_lower = np.sqrt(2 * q_e * (energy_lower - energy_scpot) / m_s)

    v_upper[v_upper < 0] = 0
    v_lower[v_lower < 0] = 0
    v_upper = np.real(v_upper)
    v_lower = np.real(v_lower)

    delta_v = v_upper - v_lower
    v_mat = np.tile(velocity, (phi_tr.shape[1], theta_tr.shape[1], 1, 1))
    v_mat = np.transpose(v_mat, [2, 3, 0, 1])

    delta_v_mat = np.tile(delta_v, (phi_tr.shape[1], theta_tr.shape[1], 1, 1))
    delta_v_mat = np.transpose(delta_v_mat, [2, 3, 0, 1])

    v_mat = v_mat[:, int_energies, ...]
    delta_v_mat = delta_v_mat[:, int_energies, ...]

    theta_mat = np.tile(theta_tr, (len(int_energies), phi_tr.shape[1], 1, 1))
    theta_mat = np.transpose(theta_mat, [2, 0, 1, 3])

    m_mat = np.sin(np.deg2rad(theta_mat)) * delta_ang

    epsilon = np.nansum(
        np.nansum(
            np.nansum(
                m_mat * vdf_diff[:, int_energies, ...] * v_mat**2 * delta_v_mat,
                axis=-1,
            ),
            axis=-1,
        ),
        axis=-1,
    )

    epsilon /= 1e6 * (n_s.data * 2)
    epsilon = ts_scalar(vdf.time.data, epsilon)

    return epsilon



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

plt.style.use("seaborn-ticks")
colors = ["tab:blue", "tab:green", "tab:red", "k"]
markers = ["s", "d", "o", "^"]


def mms_pl_config(r_mms):
    r"""Plot spacecraft configuaration with three 2d plots of the position in
    Re and one 3d plot of the relative position of the spacecraft:

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the spacecraft position

    Returns
    -------
    fig : matplotlib.pyplot.figure
        Figure with MMS configuration plot.
    axs : matplotlib.pyplot.subplotsaxes
        Axes in the figure.

    """

    r_earth = 6378.136
    r_xyz = np.vstack([np.mean(r_xyz, 0) for r_xyz in r_mms])
    r_xyz = np.mean(r_xyz, 0)
    delta_r = r_xyz - np.tile(r_xyz, (4, 1))

    fig = plt.figure(figsize=(9, 9))
    gs0 = fig.add_gridspec(
        3,
        3,
        hspace=0.3,
        left=0.1,
        right=0.9,
        bottom=0.1,
        top=0.9,
    )

    gs00 = gs0[0, :].subgridspec(1, 3, wspace=0.35)
    gs10 = gs0[1:, :].subgridspec(1, 1, wspace=0.35)

    axs0 = fig.add_subplot(gs00[0])
    axs1 = fig.add_subplot(gs00[1])
    axs2 = fig.add_subplot(gs00[2])
    axs3 = fig.add_subplot(gs10[0], projection="3d")

    earth = plt.Circle((0, 0), 1, color="k", clip_on=False)

    x_lbs = ["$X$ [$r_earth$]", "$Y$ [$r_earth$]", "$X$ [$r_earth$]"]
    y_lbs = ["$Z$ [$r_earth$]", "$Z$ [$r_earth$]", "$Y$ [$r_earth$]"]

    axs_ = [axs0, axs1, axs2]
    idxs_, idys_ = [[0, 1, 0], [2, 1, 1]]

    for ax, idx_, idy_, x_lb, y_lb in zip(axs_, idxs_, idys_, x_lbs, y_lbs):
        for i, marker in enumerate(markers):
            ax.scatter(
                r_xyz[i, idx_] / r_earth,
                r_xyz[i, idy_] / r_earth,
                marker=marker,
            )

        ax.add_artist(earth)
        ax.set_xlim([20, -20])
        ax.set_ylim([-20, 20])
        ax.set_aspect("equal")
        ax.set_xlabel(x_lb)
        ax.set_ylabel(y_lb)

    axs3.view_init(elev=13, azim=-20)

    for i, marker in enumerate(markers):
        options = {"s": 50, "marker": marker}
        axs3.scatter(delta_r[i, 0], delta_r[i, 1], delta_r[i, 2], **options)

        options = {"color": colors[i], "marker": marker, "zdir": "z", "zs": -30}
        axs3.plot([delta_r[i, 0]] * 2, [delta_r[i, 1]] * 2, **options)
        axs3.plot([delta_r[i, 0]] * 2, [delta_r[i, 2]] * 2, **options)
        axs3.plot([delta_r[i, 1]] * 2, [delta_r[i, 2]] * 2, **options)

        options = {"color": "k", "linestyle": "--", "linewidth": 0.5}
        axs3.plot(
            [delta_r[i, 0]] * 2,
            [delta_r[i, 1]] * 2,
            [-30, delta_r[i, 2]],
            **options,
        )
        axs3.plot(
            [delta_r[i, 0]] * 2,
            [-30, delta_r[i, 1]],
            [delta_r[i, 2]] * 2,
            **options,
        )
        axs3.plot(
            [-30, delta_r[i, 0]],
            [delta_r[i, 1]] * 2,
            [delta_r[i, 2]] * 2,
            **options,
        )

    for idx_0, idx_1 in zip([0, 1, 2, 0, 1, 2], [1, 2, 0, 3, 3, 3]):
        axs3.plot(
            delta_r[[idx_0, idx_1], 0],
            delta_r[[idx_0, idx_1], 1],
            delta_r[[idx_0, idx_1], 2],
            "k-",
        )

    axs3.set_xlim([-30, 30])
    axs3.set_ylim([30, -30])
    axs3.set_zlim([-30, 30])
    axs3.set_xlabel("$\\Delta X$ [km]")
    axs3.set_ylabel("$\\Delta Y$ [km]")
    axs3.set_zlabel("$\\Delta Z$ [km]")

    axs3.legend(["MMS1", "MMS2", "MMS3", "MMS4"], frameon=False)

    axs = [axs0, axs1, axs2, axs3]

    return fig, axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from .dot import dot

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def dec_par_perp(inp, b_bgd, flag_spin_plane: bool = False):
    r"""Decomposes a vector into par/perp to B components. If flagspinplane
    decomposes components to the projection of ``b0`` into the XY plane.
    ``alpha`` gives the angle between ``b0`` and the XY. plane.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the field to decompose.
    b_bgd : xarray.DataArray
        Time series of the background magnetic field.
    flag_spin_plane : bool, Optional
        Flag if True gives the projection in XY plane.

    Returns
    -------
    a_para : xarray.DataArray
        Time series of the input field parallel to the background magnetic
        field.
    a_perp : xarray.DataArray
        Time series of the input field perpendicular to the background
        magnetic field.
    alpha : xarray.DataArray
        Time series of the angle between the background magnetic field and
        the XY plane.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field (FGM) and electric field (EDP)

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Decompose e_xyz into parallel and perpendicular to b_xyz components

    >>> e_para, e_perp, _ = pyrf.dec_par_perp(e_xyz, b_xyz)

    """

    # Check arguments types
    assert isinstance(inp, xr.DataArray), "inp must be an xarray.DataArray"
    assert isinstance(b_bgd, xr.DataArray), "b_bgd must be an xarray.DataArray"
    assert isinstance(flag_spin_plane, bool), "flag_spin_plane must be boolean"

    # Check inp and b_bgd shapes
    assert inp.ndim == 2 and inp.shape[1], "inp must be a vector"
    assert b_bgd.ndim == 2 and b_bgd.shape[1], "b_bgd must be a vector"

    if not flag_spin_plane:
        b_mag = np.linalg.norm(b_bgd, axis=1, keepdims=True)

        indices = np.where(b_mag < 1e-3)[0]

        if indices.size > 0:
            b_mag[indices] = np.ones((len(indices), 1)) * 1e-3

        b_hat = b_bgd / b_mag
        b_hat = resample(b_hat, inp)

        a_para = dot(b_hat, inp)
        a_perp = inp - (b_hat.data * np.tile(a_para.data[:, np.newaxis], (1, 3)))
        alpha = []
    else:
        b_bgd = resample(b_bgd, inp)
        b_tot = np.sqrt(b_bgd[:, 0] ** 2 + b_bgd[:, 1] ** 2)
        b_bgd /= b_tot.data[:, np.newaxis]

        a_para = inp[:, 0] * b_bgd[:, 0] + inp[:, 1] * b_bgd[:, 1]
        a_perp = inp[:, 0] * b_bgd[:, 1] - inp[:, 1] * b_bgd[:, 0]
        alpha = np.arctan2(b_bgd[:, 2], b_tot)

    return a_para, a_perp, alpha



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.dates as mdates
import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pl_tx(axis, inp_list, comp, **kwargs):
    r"""Line plot of 4 spacecraft time series.

    Parameters
    ----------
    axis : matplotlib.axes._subplots.AxesSubplot
        Axis
    inp_list : list of xarray.DataArray
        Time series to plot
    comp: int
        Index of the column to plot.

    Other Parameters
    ----------------
    kwargs : dict
        Hash table of plot options.

    """

    if axis is None:
        _, axis = plt.subplots(1)

    colors = ["blue", "green", "red", "k"]

    for inp, color in zip(inp_list, colors):
        if len(inp.shape) == 3:
            data = np.reshape(
                inp.data,
                (inp.shape[0], inp.shape[1] * inp.shape[2]),
            )
        elif len(inp.shape) == 1:
            data = inp.data[:, np.newaxis]
        else:
            data = inp.data

        time = inp.time
        axis.plot(time, data[:, comp], color=color, **kwargs)

    locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
    formatter = mdates.ConciseDateFormatter(locator)
    axis.xaxis.set_major_locator(locator)
    axis.xaxis.set_major_formatter(formatter)
    axis.grid(True, which="major", linestyle="-", linewidth="0.5", c="0.5")

    return axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

# Built-in imports
import os

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


ALL_PARAMS_SCALARS = [
    "ni",
    "nbgi",
    "pbgi",
    "partni",
    "ne",
    "pbge",
    "nbge",
    "partne",
    "tsi",
    "tperpi",
    "tparai",
    "parttperpi",
    "parttparai",
    "tse",
    "tperpe",
    "tparae",
    "parttperpe",
    "parttparae",
    "pde",
    "pdi",
    "pderre",
    "pderri",
    "v",
    "v6",
    "defi",
    "defbgi",
    "defe",
    "defbge",
    "energyi",
    "bnergye",
    "epar",
    "sdev12",
    "sdev34",
    "flux-amb-pm2",
    "padlowene",
    "padmidene",
    "padhighene",
    "bpsd",
    "epsd",
]

ALL_PARAMS_VECTORS = [
    "r",
    "sti",
    "vi",
    "errvi",
    "partvi",
    "ste",
    "ve",
    "errve",
    "partve",
    "qi",
    "errqi",
    "qe",
    "errqe",
    "b",
    "e",
    "hmfe",
    "e2d",
    "es12",
    "es34",
]

ALL_PARAMS_TENSORS = [
    "pi",
    "partpi",
    "pe",
    "partpe",
    "ti",
    "partti",
    "te",
    "partte",
]

HPCA_PARAMS_SCALARS = [
    "nhplus",
    "nheplus",
    "nheplusplus",
    "noplus",
    "tshplus",
    "tsheplus",
    "tsheplusplus",
    "tsoplus",
    "dpfhplus",
    "dpfheplus",
    "dpfheplusplus",
    "dpfoplus",
    "phase",
    "adcoff",
    "saz",
    "azimuth",
]

HPCA_PARAMS_TENSORS = [
    "vhplus",
    "vheplus",
    "vheplusplus",
    "voplus",
    "phplus",
    "pheplus",
    "pheplusplus",
    "poplus",
    "thplus",
    "theplus",
    "theplusplus",
    "toplus",
]

PARAMS_SCALARS = [*ALL_PARAMS_SCALARS, *HPCA_PARAMS_SCALARS]
PARAMS_VECTORS = [*ALL_PARAMS_VECTORS]
PARAMS_TENSORS = [*ALL_PARAMS_TENSORS, *HPCA_PARAMS_TENSORS]
ALL_PARAMETERS = [*PARAMS_SCALARS, *PARAMS_VECTORS, *PARAMS_TENSORS]


COORDINATE_SYSTEMS = [
    "gse",
    "gsm",
    "dsl",
    "dbcs",
    "dmpa",
    "ssc",
    "bcs",
    "par",
]

INSTRUMENTS = [
    "mec",
    "fpi",
    "edp",
    "edi",
    "hpca",
    "fgm",
    "dfg",
    "afg",
    "scm",
    "fsm",
    "dsp",
]


SAMPLING_RATES = ["brst", "fast", "slow", "srvy"]

DATA_LVLS = ["ql", "sitl", "l1b", "l2a", "l2pre", "l2", "l3"]


def _tensor_order(splitted_key):
    param = splitted_key[0].lower()
    assert param in ALL_PARAMETERS, f"invalid PARAM : {param}"

    if param in PARAMS_SCALARS:
        tensor_order = 0
    elif param in ALL_PARAMS_VECTORS:
        tensor_order = 1
    else:
        tensor_order = 2

    return tensor_order


def tokenize(var_str):
    r"""Parses the variable keys.

    Parameters
    ----------
    var_str : str
        Input key of variable

    Returns
    -------
    out : dict
        Hash table containing :
            * param : Variable key.
            * to : Tensor order.
            * cs : Coordinate system.
            * inst : Instrument.
            * tmmode : Time mode.
            * lev" : Level of data.

    """

    splitted_key = var_str.split("_")
    assert len(splitted_key) == 4 or len(splitted_key) == 5

    tensor_order = _tensor_order(splitted_key)

    if len(splitted_key) == 5 and tensor_order > 0:
        parameter, coordinates_system, instrument, data_rate, data_level = splitted_key
        assert coordinates_system in COORDINATE_SYSTEMS, "invalid coord. sys."
    else:
        parameter, instrument, data_rate, data_level = splitted_key
        coordinates_system = []

    assert parameter in ALL_PARAMETERS, "invalid parameter"

    # Instrument
    assert instrument in INSTRUMENTS, "invalid INSTRUMENT"

    # Sampling rate
    assert data_rate in SAMPLING_RATES, "invalid sampling mode"

    # Data level
    assert data_level in DATA_LVLS, "invalid DATA_LEVEL level"

    root_path = os.path.dirname(os.path.abspath(__file__))

    with open(
        os.sep.join([root_path, "mms_keys.json"]), "r", encoding="utf-8"
    ) as json_file:
        keys_ = json.load(json_file)

    res = {
        "param": splitted_key[0],
        "to": tensor_order,
        "cs": coordinates_system,
        "inst": instrument,
        "tmmode": data_rate,
        "lev": data_level,
        "dtype": keys_[instrument][var_str.lower()]["dtype"],
        "cdf_name": keys_[instrument][var_str.lower()]["cdf_name"],
    }

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from cdflib import cdfepoch

# local imports
from .datetime642iso8601 import datetime642iso8601

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def datetime642ttns(time):
    r"""Converts datetime64 in ns units to epoch_tt2000
    (nanoseconds since J2000).

    Parameters
    ----------
    time : ndarray
        Times in datetime64 format.

    Returns
    -------
    time_ttns : ndarray
        Times in epoch_tt2000 format (nanoseconds since J2000).

    """

    # Convert to datetime64 in ns units
    time_iso8601 = datetime642iso8601(time)

    # Convert to ttns
    time_ttns = np.array([cdfepoch.parse(t_) for t_ in time_iso8601])

    return time_ttns



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def hpca_spin_sum(inp, saz, method: str = "mean"):
    r"""Sum or average teh Hot Plasma Composition Analyser (HPCA) data over
    each spin.

    Parameters
    ----------
    inp : xarray.DataArray
        Ion PSD or flux; [nt, npo16, ner63], looking direction
    saz : xarray.DataArray
        Start azimuthal spin indices.
    method : {"mean", "sum"}, Optional
        Method either "sum" or "mean". Default is "mean"

    Returns
    -------
    out : xarray.DataArray
        Distribution averaged over each spin.

    """

    az_times, start_az = [saz.time.data, saz.data]

    spin_starts = np.squeeze(np.argwhere(start_az == 0))

    out_data = []
    for i, spin in enumerate(spin_starts[:-1]):
        if method == "mean":
            out_data.append(
                inp[spin : spin_starts[i + 1]].mean(dim="time").data,
            )
        elif method == "sum":
            out_data.append(
                inp[spin : spin_starts[i + 1]].sum(dim="time").data,
            )
        else:
            raise ValueError("Invalid method")

    out_time = np.stack(az_times[spin_starts[:-1]])
    out_data = np.stack(out_data)
    coords = [inp.coords[k].data for k in inp.dims[1:]]
    coords = [out_time, *coords]
    dims = list(inp.dims)

    out = xr.DataArray(out_data, coords=coords, dims=dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_surf(axis, x, y, z, c, cmap, norm, cax_pos: str = "bottom"):
    r"""Plots surface.

    Parameters
    ----------
    axis : matplotlib.axes._subplots.Axes3DSubplot
        Axis to plot.
    x : ndarray
        X axis meshgrid.
    y : ndarray
        Y axis meshgrid.
    z : ndarray
        Z axis meshgrid.
    c : ndarray
        C axis meshgrid.
    cmap : matplotlib.colors.ListedColormap
        Colormap
    norm : matplotlib.colors.Normalize
        Normalization.
    cax_pos : {"bottom", "top", "left", "right"}, Optional
        Position of the colorbar with respect to the axis. Default is "bottom".

    Returns
    -------
    axis : matplotlib.axes._subplots.Axes3DSubplot
        Axis with surface.
    caxis : matplotlib.axes._axes.Axes
        Colorbar axis.

    """

    axis.plot_surface(x, y, z, facecolors=cmap(norm(c)))

    mappable = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    mappable.set_array(np.array([]))

    f = plt.gcf()
    pos = axis.get_position()

    if cax_pos == "bottom":
        caxis = f.add_axes([pos.x0, pos.y0 - 0.01, pos.width, 0.01])
        f.colorbar(
            mappable=mappable,
            cax=caxis,
            ax=axis,
            orientation="horizontal",
        )
    elif cax_pos == "top":
        caxis = f.add_axes(
            [pos.x0, pos.y0 + pos.height + 0.01, pos.width, 0.01],
        )
        f.colorbar(
            mappable=mappable,
            cax=caxis,
            ax=axis,
            orientation="horizontal",
        )
        caxis.xaxis.set_ticks_position("top")
        caxis.xaxis.set_label_position("top")

    elif cax_pos == "left":
        caxis = f.add_axes([pos.x0 - 0.01, pos.y0, 0.01, pos.height])
        f.colorbar(mappable=mappable, cax=caxis, ax=axis)
        caxis.yaxis.set_ticks_position("left")
        caxis.yaxis.set_label_position("left")

    elif cax_pos == "right":
        caxis = f.add_axes(
            [pos.x0 + pos.width + 0.01, pos.y0, 0.01, pos.height],
        )
        f.colorbar(mappable=mappable, cax=caxis, ax=axis)
    else:
        raise ValueError("Invalid caxis position")

    return axis, caxis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ts_spectr(time, ener, data, comp_name: str = "energy", attrs: dict = None):
    r"""Create a time series containing a spectrum

    Parameters
    ----------
    time : numpy.ndarray
        Array of times.
    ener : numpy.ndarray
        Y value of the spectrum (energies, frequencies, etc.)
    data : numpy.ndarray
        Data of the spectrum.
    attrs : dict, Optional
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        Time series of a spectrum

    """

    # Check input type
    assert isinstance(time, np.ndarray), "time must be a numpy.ndarray"
    assert isinstance(ener, np.ndarray), "time must be a numpy.ndarray"
    assert isinstance(data, np.ndarray), "data must be a numpy.ndarray"

    # Check input shape must be (n, m, )
    assert data.ndim == 2, "Input must be a spectrum"
    assert len(time) == data.shape[0], "len(time) and data.shape[0] must be equal"
    assert len(ener) == data.shape[1], "len(ener) and data.shape[1] must be equal"

    if attrs is None or not isinstance(attrs, dict):
        attrs = {}

    out = xr.DataArray(data, coords=[time, ener], dims=["time", comp_name], attrs=attrs)
    out.attrs["TENSOR_ORDER"] = 0

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party import
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def end(inp):
    """Gives the last time of the time series in unix format.

    Parameters
    ----------
    inp : xarray.DataArray or xarray.Dataset
        Time series of the input variable.

    Returns
    -------
    out : float
        Value of the last time in unix format.

    """

    message = "inp must be a xarray.DataArray or xarray.Dataset"
    assert isinstance(inp, (xr.DataArray, xr.Dataset)), message

    out = inp.time.data[-1].astype(np.int64) / 1e9

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def remove_idist_background(vdf, def_bg):
    r"""Remove the mode background population due to penetrating radiation
    `def_bg` from the ion velocity distribution function `vdf` using the
    method from [1]_.

    Parameters
    ----------
    vdf : xarray.Dataset
        Ion velocity distribution function.
    def_bg : xarray.DataArray
        Omni-directional ion differential energy flux.

    Returns
    -------
    vdf_new : xarray.Dataset
        Ion velocity distribution function cleaned.

    References
    ----------
    .. [1]  Gershman, D. J., Dorelli, J. C., Avanov,L. A., Gliese, U.,
            Barrie, A., Schiff, C.,et al. (2019). Systematic uncertainties
            in plasma parameters reported by the fast plasma investigation on
            NASA's magnetospheric multiscale mission. Journal of Geophysical
            Research: Space Physics, 124, https://doi.org/10.1029/2019JA026980

    """

    # Tile background flux to number of energy channels of the
    # FPI-DIS instrument
    def_bg_tmp = np.tile(def_bg.data[:, np.newaxis], (1, vdf.energy.shape[1]))

    # Convert differential energy flux (cm^2 s sr)^{-1} of the background
    # population (penetrating radiations) to phase-space density (s^3 m^{-6})
    coeff = constants.proton_mass / (constants.elementary_charge * vdf.energy.data)
    vdf_bg = def_bg_tmp.copy() * 1e4 / 2
    vdf_bg *= coeff**2
    vdf_bg /= 1e12

    # Tile the background phase-space density to number of azimuthal
    # and elevation angles channels of the FPI-DIS instrument
    vdf_bg = np.tile(
        vdf_bg[:, :, np.newaxis, np.newaxis],
        (1, 1, vdf.phi.shape[1], vdf.theta.shape[0]),
    )

    vdf_new = vdf.copy()
    vdf_new.data.data -= vdf_bg.data

    return vdf_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import re

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def iso86012timevec(time):
    r"""Convert ISO 8601 time string into time vector.

    Parameters
    ----------
    time : ndarray or list or str
        Time in ISO 8601 format YYYY-MM-DDThh:mm:ss.mmmuuunnn.

    Returns
    -------
    time_vec : list
        Time vector.

    See Also
    --------
    pyrfu.pyrf.iso86012timevec

    """

    iso_8601 = (
        r"(?P<years>[0-9]{4})-(?P<months>[0-9]{2})-(?P<days>[0-9]{2})"
        r"T(?P<hours>[0-9]{2}):(?P<minutes>[0-9]{2})"
        r":(?P<seconds>[0-9]{2}).(?P<miliseconds>[0-9]{3})"
        r"(?P<microseconds>[0-9]{3})(?P<nanoseconds>[0-9]{3})"
    )

    # Define parser
    fmt = re.compile(iso_8601)

    # Make time is a 1d array
    time = np.atleast_1d(time)

    time_vec = [[int(p_) for p_ in fmt.match(t_).groups()] for t_ in time]
    time_vec = np.array(time_vec)

    return time_vec



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def cross(inp1, inp2):
    r"""Computes cross product of two fields.

    Parameters
    ----------
    inp1 : xarray.DataArray
        Time series of the first field X.
    inp2 : xarray.DataArray
        Time series of the second field Y.

    Returns
    -------
    out : xarray.DataArray
        Time series of the cross product Z = XxY.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Define time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Index of the MMS spacecraft

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)

    Compute magnitude of the magnetic field

    >>> b_mag = pyrf.norm(b_xyz)

    Compute ExB drift velocity

    >>> v_xyz_exb = pyrf.cross(e_xyz, b_xyz) / b_mag ** 2

    """

    # Check type
    assert isinstance(inp1, xr.DataArray), "inp1 must be a xarray.DataArray"
    assert isinstance(inp2, xr.DataArray), "inp2 must be a xarray.DataArray"

    # Check inputs are vectors
    assert inp1.ndim == 2 and inp1.shape[1] == 3, "inp1 must be a vector"
    assert inp2.ndim == 2 and inp2.shape[1] == 3, "inp1 must be a vector"

    if len(inp1) != len(inp2):
        inp2 = resample(inp2, inp1)

    out_data = np.cross(inp1.data, inp2.data, axis=1)

    out = ts_vec_xyz(inp1.time.data, out_data)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from cdflib import cdfepoch, cdfread
from dateutil import parser

# Local imports
from .datetime2iso8601 import datetime2iso8601

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def read_cdf(path, tint):
    r"""Reads CDF files.

    Parameters
    ----------
    path : str
        String of the filename in .cdf containing the L2 data
    tint : list
        Time interval

    Returns
    -------
    out_dict : dict
        Hash table with fields contained in the .cdf file.

    """

    tint = list(map(parser.parse, tint))
    tint = list(map(datetime2iso8601, tint))
    tint = list(map(cdfepoch.parse, tint))

    out_dict = {}

    with cdfread.CDF(path) as file:
        keys_ = file.cdf_info()["zVariables"]
        for k_ in keys_:
            temp_ = file.varget(k_, starttime=tint[0], endtime=tint[1])
            shape_ = temp_.shape
            coords = [np.arange(lim_) for lim_ in shape_]

            out_dict[k_.lower()] = xr.DataArray(temp_, coords=coords)

    return out_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _get_vol_ten(r_xyz, time):
    if len(time) == 1:
        time = np.array([time, time, time, time])

    tckr_x, tckr_y, tckr_z = [[], [], []]

    for i in range(4):
        tckr_x.append(
            interpolate.interp1d(r_xyz[i].time.data, r_xyz[i].data[:, 0]),
        )
        tckr_y.append(
            interpolate.interp1d(r_xyz[i].time.data, r_xyz[i].data[:, 1]),
        )
        tckr_z.append(
            interpolate.interp1d(r_xyz[i].time.data, r_xyz[i].data[:, 2]),
        )

        r_xyz[i] = np.array(
            [tckr_x[i](time[0]), tckr_y[i](time[0]), tckr_z[i](time[0])],
        )

    # Volumetric tensor with SC1 as center.
    dr_mat = (np.vstack(r_xyz[1:]) - np.tile(r_xyz[0], (3, 1))).T

    return dr_mat


def c_4_v(r_xyz, time):
    r"""Calculates velocity or time shift of discontinuity as in [6]_.

    Parameters
    ----------
    r_xyz : list
        Time series of the positions of the 4 spacecraft.
    time : list
        Crossing times or time and velocity.

    Returns
    -------
    out : ndarray
        Discontinuity velocity or time shift with respect to mms1.

    References
    ----------
    .. [6]	Vogt, J., Haaland, S., and Paschmann, G. (2011) Accuracy
            of multi-point boundary crossing time analysis, Ann.
            Geophys., 29, 2239-2252, doi :
            https://doi.org/10.5194/angeo-29-2239-2011

    """

    if isinstance(time, np.ndarray) and time.dtype == np.datetime64:
        flag = "v_from_t"

        time = time.view("i8") * 1e-9
    elif time[1] > 299792.458:
        flag = "v_from_t"
    else:
        flag = "dt_from_v"

    if flag.lower() == "v_from_t":
        # Time input, velocity output
        dr_mat = _get_vol_ten(r_xyz, time)
        tau = np.array(time[1:]) - time[0]
        slowness = np.linalg.solve(dr_mat, tau)

        # "1/v vector"
        out = slowness / np.linalg.norm(slowness) ** 2

    elif flag.lower() == "dt_from_v":
        # Time and velocity input, time output
        time_center = time[0]  # center time
        velocity = np.array(time[1:])  # Input velocity
        slowness = velocity / np.linalg.norm(velocity) ** 2

        dr_mat = _get_vol_ten(r_xyz, time_center)

        delta_t = np.matmul(dr_mat, slowness)
        out = np.hstack([0, delta_t])

    else:
        raise ValueError("Invalid flag")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ts_skymap(time, data, energy, phi, theta, **kwargs):
    r"""Creates a skymap of the distribution function.

    Parameters
    ----------
    time : array_like
        List of times.
    data : array_like
        Values of the distribution function.
    energy : array_like
        Energy levels.
    phi : array_like
        Azimuthal angles.
    theta : array_like
        Elevation angles.

    Other Parameters
    ---------------
    **kwargs
        Hash table of keyword arguments with :
            * energy0 : array_like
                Energy table 0 (odd time indices).
            * energy1 : array_like
                Energy table 1 (even time indices).
            * esteptable : array_like
                Time series of the stepping table between energies (burst).

    Returns
    -------
    out : xarray.Dataset
        Skymap of the distribution function.

    """

    # Check input type
    assert isinstance(time, np.ndarray), "time must be numpy.ndarray"
    assert isinstance(data, np.ndarray), "data must be numpy.ndarray"
    assert isinstance(energy, np.ndarray), "energy must be numpy.ndarray"
    assert isinstance(phi, np.ndarray), "phi must be numpy.ndarray"
    assert isinstance(theta, np.ndarray), "theta must be numpy.ndarray"

    # Check if even (odd) time step energy channels energy1 (energy0), and
    # energy step table are provided.
    energy0 = kwargs.get("energy0", energy[0, :])
    energy1 = kwargs.get("energy1", energy[1, :])
    esteptable = kwargs.get("esteptable", np.zeros(len(time)))

    # Check that energy0 and energy1
    # assert isinstance(energy0, np.ndarray), "energy0 must be 1D numpy.ndarray"
    # assert energy0.ndim == 1, "energy0 must be 1D numpy.ndarray"
    # assert energy0.shape[0] == energy.shape[1], "energy0 is not consistent with time"
    # assert isinstance(energy1, np.ndarray), "energy1 must be 1D numpy.ndarray"
    # assert energy1.ndim == 1, "energy1 must be 1D numpy.ndarray"
    # assert energy1.shape[0] == energy.shape[1], "energy1 is not consistent with time"

    # Check esteptable
    assert isinstance(esteptable, np.ndarray), "esteptable must be 1D numpy.ndarray"
    assert esteptable.ndim == 1, "esteptable must be 1D numpy.ndarray"
    assert esteptable.shape[0] == len(time), "esteptable is not consistent with time"

    attrs = kwargs.get("attrs", {})
    coords_attrs = kwargs.get("coords_attrs", {})
    glob_attrs = kwargs.get("glob_attrs", {})

    # Check attributes are dictionaries
    assert isinstance(attrs, dict)

    out_dict = {
        "data": (["time", "idx0", "idx1", "idx2"], data),
        "phi": (["time", "idx1"], phi),
        "theta": (["idx2"], theta),
        "energy": (["time", "idx0"], energy),
        "time": time,
        "idx0": np.arange(energy.shape[1]),
        "idx1": np.arange(phi.shape[1]),
        "idx2": np.arange(len(theta)),
    }

    # Construct global attributes and sort them
    # remove energy0, energy1, and esteptable from global attrs to overwrite
    overwrite_keys = ["energy0", "energy1", "esteptable"]
    glob_attrs = {k: glob_attrs[k] for k in glob_attrs if k not in overwrite_keys}
    glob_attrs = {
        "energy0": energy0,
        "energy1": energy1,
        "esteptable": esteptable,
        **glob_attrs,
    }

    glob_attrs = {k: glob_attrs[k] for k in sorted(glob_attrs)}

    # Create Dataset
    out = xr.Dataset(out_dict, attrs=glob_attrs)

    # Sort and fill coordinates attributes
    for k in coords_attrs:
        out[k].attrs = {m: coords_attrs[k][m] for m in sorted(coords_attrs[k])}

    # Sort and fill data attributes
    out.data.attrs = {k: attrs[k] for k in sorted(attrs)}

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .calculate_epsilon import calculate_epsilon
from .copy_files import copy_files
from .copy_files_ancillary import copy_files_ancillary
from .correct_edp_probe_timing import correct_edp_probe_timing
from .db_get_ts import db_get_ts
from .db_get_variable import db_get_variable
from .db_init import db_init
from .def2psd import def2psd
from .dft_time_shift import dft_time_shift
from .download_ancillary import download_ancillary
from .download_data import download_data
from .dpf2psd import dpf2psd
from .dsl2gse import dsl2gse
from .dsl2gsm import dsl2gsm

# Energetic Ion Spectrometer (EIS)
from .eis_ang_ang import eis_ang_ang
from .eis_combine_proton_pad import eis_combine_proton_pad
from .eis_combine_proton_skymap import eis_combine_proton_skymap
from .eis_combine_proton_spec import eis_combine_proton_spec

# from .eis_correlation import eis_correlation
from .eis_moments import eis_moments
from .eis_omni import eis_omni
from .eis_pad import eis_pad
from .eis_pad_combine_sc import eis_pad_combine_sc
from .eis_pad_spinavg import eis_pad_spinavg
from .eis_proton_correction import eis_proton_correction
from .eis_skymap import eis_skymap
from .eis_skymap_combine_sc import eis_skymap_combine_sc
from .eis_spec_combine_sc import eis_spec_combine_sc
from .eis_spin_avg import eis_spin_avg
from .estimate_phase_speed import estimate_phase_speed

# Fly’s Eye Energetic Particle Spectrometer (FEEPS)
from .feeps_active_eyes import feeps_active_eyes
from .feeps_correct_energies import feeps_correct_energies
from .feeps_energy_table import feeps_energy_table
from .feeps_flat_field_corrections import feeps_flat_field_corrections
from .feeps_omni import feeps_omni
from .feeps_pad import feeps_pad
from .feeps_pad_spinavg import feeps_pad_spinavg
from .feeps_pitch_angles import feeps_pitch_angles
from .feeps_remove_bad_data import feeps_remove_bad_data
from .feeps_remove_sun import feeps_remove_sun
from .feeps_sector_spec import feeps_sector_spec
from .feeps_spin_avg import feeps_spin_avg
from .feeps_split_integral_ch import feeps_split_integral_ch
from .fft_bandpass import fft_bandpass
from .fk_power_spectrum_4sc import fk_power_spectrum_4sc
from .get_data import get_data
from .get_dist import get_dist
from .get_eis_allt import get_eis_allt
from .get_feeps_alleyes import get_feeps_alleyes
from .get_feeps_omni import get_feeps_omni

# Hot Plasma Composition Analyser (HPCA)
from .get_hpca_dist import get_hpca_dist
from .get_pitch_angle_dist import get_pitch_angle_dist
from .get_ts import get_ts
from .get_variable import get_variable
from .hpca_calc_anodes import hpca_calc_anodes
from .hpca_energies import hpca_energies
from .hpca_pad import hpca_pad
from .hpca_spin_sum import hpca_spin_sum
from .lh_wave_analysis import lh_wave_analysis
from .list_files import list_files
from .list_files_ancillary import list_files_ancillary
from .list_files_ancillary_sdc import list_files_ancillary_sdc
from .list_files_sdc import list_files_sdc
from .load_ancillary import load_ancillary
from .load_brst_segments import load_brst_segments
from .make_model_kappa import make_model_kappa
from .make_model_vdf import make_model_vdf
from .probe_align_times import probe_align_times
from .psd2def import psd2def
from .psd2dpf import psd2dpf
from .psd_moments import psd_moments
from .psd_rebin import psd_rebin
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv
from .reduce import reduce
from .remove_edist_background import remove_edist_background
from .remove_idist_background import remove_idist_background
from .remove_imoms_background import remove_imoms_background
from .rotate_tensor import rotate_tensor
from .scpot2ne import scpot2ne
from .spectr_to_dataset import spectr_to_dataset

# @Louis Richard
from .tokenize import tokenize
from .vdf_elim import vdf_elim
from .vdf_omni import vdf_omni
from .vdf_projection import vdf_projection
from .vdf_reduce import vdf_frame_transformation, vdf_reduce
from .vdf_to_e64 import vdf_to_e64
from .whistler_b2e import whistler_b2e

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = [
    "calculate_epsilon",
    "copy_files",
    "copy_files_ancillary",
    "correct_edp_probe_timing",
    "db_get_ts",
    "db_get_variable",
    "db_init",
    "def2psd",
    "dft_time_shift",
    "download_ancillary",
    "download_data",
    "dpf2psd",
    "dsl2gse",
    "dsl2gsm",
    "eis_ang_ang",
    "eis_combine_proton_pad",
    "eis_combine_proton_skymap",
    "eis_combine_proton_spec",
    "eis_moments",
    "eis_omni",
    "eis_pad",
    "eis_pad_combine_sc",
    "eis_pad_spinavg",
    "eis_proton_correction",
    "eis_skymap",
    "eis_skymap_combine_sc",
    "eis_spec_combine_sc",
    "eis_spin_avg",
    "estimate_phase_speed",
    "feeps_active_eyes",
    "feeps_correct_energies",
    "feeps_energy_table",
    "feeps_flat_field_corrections",
    "feeps_omni",
    "feeps_pad",
    "feeps_pad_spinavg",
    "feeps_pitch_angles",
    "feeps_remove_bad_data",
    "feeps_remove_sun",
    "feeps_sector_spec",
    "feeps_spin_avg",
    "feeps_split_integral_ch",
    "fft_bandpass",
    "fk_power_spectrum_4sc",
    "get_data",
    "get_dist",
    "get_eis_allt",
    "get_feeps_alleyes",
    "get_feeps_omni",
    "get_hpca_dist",
    "get_pitch_angle_dist",
    "get_ts",
    "get_variable",
    "hpca_calc_anodes",
    "hpca_energies",
    "hpca_pad",
    "hpca_spin_sum",
    "lh_wave_analysis",
    "list_files",
    "list_files_sdc",
    "list_files_ancillary",
    "list_files_ancillary_sdc",
    "load_ancillary",
    "load_brst_segments",
    "make_model_kappa",
    "make_model_vdf",
    "probe_align_times",
    "psd2def",
    "psd2dpf",
    "psd_moments",
    "psd_rebin",
    "read_feeps_sector_masks_csv",
    "reduce",
    "remove_edist_background",
    "remove_idist_background",
    "remove_imoms_background",
    "rotate_tensor",
    "scpot2ne",
    "spectr_to_dataset",
    "tokenize",
    "vdf_elim",
    "vdf_frame_transformation",
    "vdf_omni",
    "vdf_projection",
    "vdf_reduce",
    "vdf_to_e64",
    "whistler_b2e",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from cdflib import cdfepoch

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _compose_date(
    years,
    months,
    days,
    hours=None,
    minutes=None,
    seconds=None,
    milliseconds=None,
    microseconds=None,
    nanoseconds=None,
):
    years = np.asarray(years) - 1970
    months = np.asarray(months) - 1
    days = np.asarray(days) - 1
    types = [
        "<M8[Y]",
        "<m8[M]",
        "<m8[D]",
        "<m8[h]",
        "<m8[m]",
        "<m8[s]",
        "<m8[ms]",
        "<m8[us]",
        "<m8[ns]",
    ]
    vals = [
        years,
        months,
        days,
        hours,
        minutes,
        seconds,
        milliseconds,
        microseconds,
        nanoseconds,
    ]

    dates_list = []
    for t, v in zip(types, vals):
        if v is not None:
            dates_list.append(np.asarray(v, dtype=t))

    dates = sum(dates_list)

    return dates


def cdfepoch2datetime64(epochs):
    r"""Converts CDF epochs to numpy.datetime64 with nanosecond precision.

    Parameters
    ----------
    epochs : float or int or array_like
        CDF epochs to convert.

    Returns
    -------
    times : numpy.ndarray
        Array of times in datetime64([ns]).

    """

    # Check input type
    times = cdfepoch.breakdown(epochs)
    times = np.transpose(np.atleast_2d(times))

    times = _compose_date(*times).astype("datetime64[ns]")

    return times
